@@ -106,10 +106,6 @@ void init_builtin(void)
     DEFBUILTIN("pwd", pwd_builtin, BI_SEMISPECIAL, pwd_help);
     DEFBUILTIN("hash", hash_builtin, BI_REGULAR, hash_help);
     DEFBUILTIN("umask", umask_builtin, BI_SEMISPECIAL, umask_help);
-#if YASH_ENABLE_DIRSTACK
-    DEFBUILTIN("pushd", pushd_builtin, BI_REGULAR, pushd_help);
-    DEFBUILTIN("popd", popd_builtin, BI_REGULAR, popd_help);
-#endif
 
     /* defined in "alias.c" */
 #if YASH_ENABLE_ALIAS
@@ -129,6 +125,8 @@ void init_builtin(void)
     DEFBUILTIN("getopts", getopts_builtin, BI_SEMISPECIAL, getopts_help);
     DEFBUILTIN("read", read_builtin, BI_SEMISPECIAL, read_help);
 #if YASH_ENABLE_DIRSTACK
+    DEFBUILTIN("pushd", pushd_builtin, BI_REGULAR, pushd_help);
+    DEFBUILTIN("popd", popd_builtin, BI_REGULAR, popd_help);
     DEFBUILTIN("dirs", dirs_builtin, BI_REGULAR, dirs_help);
 #endif
 
@@ -283,13 +281,9 @@ int help_builtin(int argc, void **argv)
     if (xoptind == argc)
 	goto print_help;
 
-    bool err = false;
-    while (xoptind < argc) {
-	if (print_builtin_help(ARGV(xoptind)) != Exit_SUCCESS)
-	    err = true;
-	xoptind++;
-    }
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    while (xoptind < argc)
+	print_builtin_help(ARGV(xoptind++));
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 }
 
 const char help_help[] = Ngt(
