@@ -35,8 +35,6 @@
 #include "ulimit.h"
 
 
-/********** Builtin **********/
-
 static const struct resource {
     wchar_t option;
     int type;
@@ -77,6 +75,7 @@ static const struct resource {
     { 0, 0, 0, NULL, },
 };
 
+/* The "ulimit" builtin */
 int ulimit_builtin(int argc, void **argv)
 {
     const wchar_t *short_options = L"HSa"
@@ -148,7 +147,9 @@ int ulimit_builtin(int argc, void **argv)
 #if HAVE_RLIMIT_LOCKS
 	{ L"locks", xno_argument, L'x', },
 #endif
+#if YASH_ENABLE_HELP
 	{ L"help", xno_argument, L'-', },
+#endif
 	{ NULL, 0, 0, },
     };
 
@@ -163,8 +164,10 @@ int ulimit_builtin(int argc, void **argv)
 	    case L'H':  type = hard;       break;
 	    case L'S':  type = soft;       break;
 	    case L'a':  print_all = true;  break;
+#if YASH_ENABLE_HELP
 	    case L'-':
 		return print_builtin_help(ARGV(0));
+#endif
 	    default:
 		for (const struct resource *r = resource_types; r->option; r++){
 		    if (r->option == opt) {
@@ -288,6 +291,7 @@ print_usage:
     return Exit_ERROR;
 }
 
+#if YASH_ENABLE_HELP
 const char ulimit_help[] = Ngt(
 "ulimit - set or print resource limitation\n"
 "\tulimit [-HSaefilnqrstuvx] [limit]\n"
@@ -295,9 +299,9 @@ const char ulimit_help[] = Ngt(
 "options listed below.\n"
 "For each resource type, there are two limit values: the hard limit and the\n"
 "soft limit. You can change a soft limit freely as long as it does not exceed\n"
-"the hard limit. You can decrease a hard limit, but cannot increase it\n"
-"without a proper permission. <limit> must be a non-negative integer or one\n"
-"of \"hard\", \"soft\" and \"unlimited\".\n"
+"the hard limit. You can decrease a hard limit but cannot increase it without\n"
+"a proper permission. <limit> must be a non-negative integer or one of\n"
+"\"hard\", \"soft\" and \"unlimited\".\n"
 "The resource limits will affect the current shell process and all\n"
 "succeedingly invoked commands.\n"
 "If <limit> is not specified, the current limit is printed.\n"
@@ -327,10 +331,11 @@ const char ulimit_help[] = Ngt(
 " -u --nproc:      The maximum number of processes for a user\n"
 " -v --as:         The maximum size of memory used by a process (K)\n"
 " -x --locks:      The maximum number of file locks\n"
-"For resource types marked with (B), the limits are set and printed in 512-\n"
-"byte blocks, (K) in kilobytes, and (S) in seconds.\n"
+"For resource types marked with (B), (K) and (S), the limits are set and\n"
+"printed in 512-byte blocks, kilobytes and seconds respectively.\n"
 "Available resource types vary depending on your system.\n"
 );
+#endif /* YASH_ENABLE_HELP */
 
 
 /* vim: set ts=8 sts=4 sw=4 noet tw=80: */
