@@ -21,17 +21,21 @@
 #ifndef YASH_COMMON_H
 #define YASH_COMMON_H
 
-#define _POSIX_C_SOURCE 200112L
-#define _XOPEN_SOURCE   600
+#if HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#ifndef _POSIX_C_SOURCE
+# define _POSIX_C_SOURCE 200112L
+#endif
+#ifndef _XOPEN_SOURCE
+# define _XOPEN_SOURCE 600
+#endif
 
 #ifndef __GNUC__
 # define __attribute__(ignore)
 #endif
 
-#if HAVE_CONFIG_H
-# include "config.h"
-#endif
-
 #if HAVE_GETTEXT
 # define gt(MSGID)  ((const char *) gettext(MSGID))
 # define Ngt(MSGID) MSGID
@@ -51,4 +55,4 @@
 #endif /* YASH_COMMON_H */
 
 
-/* vim: set ts=8 sts=4 sw=4 noet: */
+/* vim: set ts=8 sts=4 sw=4 noet tw=80: */
