@@ -1109,6 +1109,7 @@ void exec_simple_command(
     case specialbuiltin:
     case semispecialbuiltin:
     case regularbuiltin:
+	yash_error_message_count = 0;
 	current_builtin_name = argv[0];  // XXX not reentrant
 	laststatus = ci->ci_builtin(argc, argv);
 	current_builtin_name = NULL;
@@ -1139,7 +1140,10 @@ void print_xtrace(void *const *argv)
     if (shopt_xtrace && (tracevars || argv != NULL)) {
 	bool first = true;
 
-	print_prompt(4);
+	struct promptset_T prompt = get_prompt(4);
+	print_prompt(prompt.main);
+	print_prompt(prompt.styler);
+
 	if (tracevars) {
 	    fprintf(stderr, "%ls", xtrace_buffer.contents + 1);
 	    first = false;
@@ -1153,6 +1157,9 @@ void print_xtrace(void *const *argv)
 	    }
 	}
 	fputc('\n', stderr);
+
+	print_prompt(PROMPT_RESET);
+	free_prompt(prompt);
     }
     if (xtrace_buffer.contents) {
 	wb_destroy(&xtrace_buffer);
@@ -1194,13 +1201,12 @@ void exec_fall_back_on_sh(
     free(objpath);
 #elif HAVE_PATHS_H && defined _PATH_BSHELL
     xexecve(_PATH_BSHELL, args, envp);
-#else
+#endif
     const char *shpath = get_command_path("sh", false);
     if (shpath)
 	xexecve(shpath, args, envp);
     else
 	errno = ENOENT;
-#endif
     xerror(errno, Ngt("cannot invoke new shell to execute `%s'"), argv[0]);
 }
 
@@ -1965,21 +1971,21 @@ int command_builtin(int argc, void **argv)
 	return command_builtin_execute(
 		argc - xoptind, argv + xoptind, type);
     } else {
-	bool err = false;
 	bool keywordalias = false;
 	if (!builtin && !external)
 	    type |= sct_function, keywordalias = true;
 
+	bool ok = true;
 	clearerr(stdout);
-	for (int i = xoptind; i < argc; i++)
-	    err |= !print_command_info(
+	for (int i = xoptind; i < argc; i++) {
+	    ok &= print_command_info(
 		    ARGV(i), type, keywordalias, humanfriendly);
-	if (!ferror(stdout)) {
-	    return err ? Exit_FAILURE : Exit_SUCCESS;
-	} else {
-	    xerror(0, Ngt("cannot print to standard output"));
-	    return Exit_FAILURE;
+	    if (ferror(stdout)) {
+		xerror(0, Ngt("cannot print to standard output"));
+		return Exit_FAILURE;
+	    }
 	}
+	return ok ? Exit_SUCCESS : Exit_FAILURE;
     }
 
 print_usage:
