@@ -42,6 +42,7 @@ typedef enum sigtype_T {
 
 extern int laststatus, savelaststatus;
 extern pid_t lastasyncpid;
+extern _Bool special_builtin_executed;
 
 struct execinfo;
 extern void reset_execinfo(void);
@@ -53,11 +54,11 @@ extern _Bool return_pending(void)
     __attribute__((pure));
 
 struct and_or_T;
-struct wordunit_T;
+struct embedcmd_T;
 extern void exec_and_or_lists(const struct and_or_T *a, _Bool finally_exit);
 extern pid_t fork_and_reset(pid_t pgid, _Bool fg, sigtype_T sigtype);
 extern struct xwcsbuf_T *get_xtrace_buffer(void);
-extern wchar_t *exec_command_substitution(const wchar_t *code)
+extern wchar_t *exec_command_substitution(const struct embedcmd_T *cmdsub)
     __attribute__((nonnull,malloc,warn_unused_result));
 extern int exec_variable_as_commands(
 	const wchar_t *varname, const char *codename)
