@@ -19,8 +19,8 @@
 #ifndef YASH_HASHTABLE_H
 #define YASH_HASHTABLE_H
 
-#include <stddef.h>
 #include <stdint.h>
+#include <stdlib.h>
 
 
 #if defined UINT64_MAX && UINT64_MAX == UINT_FAST32_MAX
@@ -43,7 +43,7 @@ typedef hashval_T hashfunc_T(const void *key);
 typedef int keycmp_T(const void *key1, const void *key2);
 
 typedef struct hashtable_T {
-    size_t count, capacity;
+    size_t capacity, count;
     hashfunc_T *hashfunc;
     keycmp_T *keycmp;
     size_t emptyindex, tailindex;
@@ -61,7 +61,9 @@ extern hashtable_T *ht_initwithcapacity(
 	hashtable_T *ht, hashfunc_T *hashfunc, keycmp_T *keycmp,
 	size_t capacity)
     __attribute__((nonnull));
-extern void ht_destroy(hashtable_T *ht)
+static inline void ht_destroy(hashtable_T *ht)
+    __attribute__((nonnull));
+extern hashtable_T *ht_setcapacity(hashtable_T *ht, size_t newcapacity)
     __attribute__((nonnull));
 extern hashtable_T *ht_ensurecapacity(hashtable_T *ht, size_t capacity)
     __attribute__((nonnull));
@@ -95,15 +97,23 @@ extern void kvfree(kvpair_T kv);
 
 #define HASHTABLE_DEFAULT_INIT_CAPACITY 5
 
-/* Initializes a hashtable with the default capacity.
- * `hashfunc' is a hash function to hash keys.
- * `keycmp' is a function that compares two keys. */
+/* Initializes the specified hashtable with the default capacity.
+ * `hashfunc' is the hash function to hash keys.
+ * `keycmp' is the function that compares two keys. */
 hashtable_T *ht_init(hashtable_T *ht, hashfunc_T *hashfunc, keycmp_T *keycmp)
 {
     return ht_initwithcapacity(
 	    ht, hashfunc, keycmp, HASHTABLE_DEFAULT_INIT_CAPACITY);
 }
 
+/* Destroys the specified hashtable.
+ * Note that this function doesn't `free' any keys or values. */
+void ht_destroy(hashtable_T *ht)
+{
+    free(ht->indices);
+    free(ht->entries);
+}
+
 
 #endif /* YASH_HASHTABLE_H */
 
