@@ -57,6 +57,7 @@ extern le_command_func_T
     cmd_bol_or_digit,
     cmd_accept_line,
     cmd_abort_line,
+    cmd_eof,
     cmd_eof_if_empty,
     cmd_eof_or_delete,
     cmd_accept_with_hash,
@@ -139,6 +140,7 @@ extern le_command_func_T
     cmd_vi_append,
     cmd_vi_append_end,
     cmd_vi_replace,
+    cmd_vi_switch_case,
     cmd_vi_switch_case_char,
     cmd_vi_yank,
     cmd_vi_yank_to_eol,
