@@ -36,12 +36,12 @@
 le_mode_T le_modes[LE_MODE_N];
 
 /* The current editing mode.
- * Points to one of `le_modes'. */
+ * Points to one of the modes in `le_modes'. */
 le_mode_T *le_current_mode;
 
 
 /* Initializes `le_modes' if not yet initialized.
- * May be called more than once. */
+ * May be called more than once but does nothing if so. */
 void le_keymap_init(void)
 {
     static bool initialized = false;
@@ -245,6 +245,8 @@ void le_keymap_init(void)
     Set(Key_c_k,            cmd_forward_kill_line);
     Set(Key_kill,           cmd_backward_kill_line);
     Set(Key_c_u,            cmd_backward_kill_line);
+    Set(Key_c_x Key_backspace, cmd_backward_kill_line);
+    Set(Key_c_x Key_erase,  cmd_backward_kill_line);
     Set(Key_c_y,            cmd_put_left);
     Set(Key_escape "y",     cmd_put_pop);
     Set(Key_escape "Y",     cmd_put_pop);
@@ -266,10 +268,10 @@ void le_keymap_init(void)
     Set(Key_escape "C",     cmd_emacs_capitalize_word);
     Set(Key_escape Key_backslash, cmd_emacs_delete_horizontal_space);
     Set(Key_escape " ",     cmd_emacs_just_one_space);
-    Set(Key_escape "<",     cmd_oldest_history_eol);
-    Set(Key_escape ">",     cmd_return_history_eol);
     Set(Key_c_s,            cmd_emacs_search_forward);
     Set(Key_c_r,            cmd_emacs_search_backward);
+    Set(Key_escape "<",     cmd_oldest_history_eol);
+    Set(Key_escape ">",     cmd_return_history_eol);
     Set(Key_down,           cmd_next_history_eol);
     Set(Key_c_n,            cmd_next_history_eol);
     Set(Key_up,             cmd_prev_history_eol);
@@ -278,7 +280,7 @@ void le_keymap_init(void)
 
     le_modes[LE_MODE_EMACS_SEARCH].default_command = cmd_srch_self_insert;
     t = trie_create();
-    Set(Key_c_v, cmd_expect_verbatim);
+    Set(Key_c_v,       cmd_expect_verbatim);
     Set(Key_c_m,       cmd_accept_line);
     Set(Key_interrupt, cmd_abort_line);
     Set(Key_c_c,       cmd_abort_line);
@@ -312,13 +314,7 @@ void le_keymap_init(void)
 void le_set_mode(le_mode_id_T id)
 {
     assert(id < LE_MODE_N);
-    le_current_mode = &le_modes[id];
-}
-
-/* Returns the current editing mode. */
-le_mode_id_T le_get_mode(void)
-{
-    return (le_mode_id_T) (le_current_mode - le_modes);
+    le_current_mode = le_id_to_mode(id);
 }
 
 
@@ -341,7 +337,7 @@ static const char *get_command_name(le_command_func_T *command)
     __attribute__((nonnull,const));
 
 /* Array of pairs of a command name and function.
- * Sorted by the name. */
+ * Sorted by name. */
 static const struct command_name_pair {
     const char *name;
     le_command_func_T *command;
@@ -361,7 +357,9 @@ int bindkey_builtin(int argc, void **argv)
 	{ L"vi-command", xno_argument, L'a', },
 	{ L"emacs",      xno_argument, L'e', },
 	{ L"list",       xno_argument, L'l', },
+#if YASH_ENABLE_HELP
 	{ L"help",       xno_argument, L'-', },
+#endif
 	{ NULL, 0, 0, },
     };
 
@@ -376,8 +374,10 @@ int bindkey_builtin(int argc, void **argv)
 	    case L'e':  mode = LE_MODE_EMACS;       break;
 	    case L'v':  mode = LE_MODE_VI_INSERT;   break;
 	    case L'l':  list = true;                break;
+#if YASH_ENABLE_HELP
 	    case L'-':
 		return print_builtin_help(ARGV(0));
+#endif
 	    default:  print_usage:
 		fprintf(stderr,
 			gt("Usage:  bindkey -aev [keyseq [command]]\n"
@@ -403,7 +403,7 @@ int bindkey_builtin(int argc, void **argv)
 
     if (xoptind == argc) {
 	/* print all key bindings */
-	le_mode_T *m = &le_modes[mode];
+	le_mode_T *m = le_id_to_mode(mode);
 	return trie_foreachw(m->keymap, print_binding_main, m);
     } else if (xoptind + 1 == argc) {
 	return print_binding(mode, ARGV(xoptind));
@@ -426,7 +426,7 @@ int print_all_commands(void)
     return Exit_SUCCESS;
 }
 
-/* Binds `keyseq' to the specified command.
+/* Binds the specified key sequence to the specified command.
  * If `commandname' is L"-", the binding is removed.
  * If the specified command is not found, it is an error. */
 int set_key_binding(
@@ -487,17 +487,12 @@ int print_binding(le_mode_id_T mode, const wchar_t *keyseq)
 {
     trieget_T tg = trie_getw(le_modes[mode].keymap, keyseq);
 
-    switch (tg.type) {
-	case TG_NOMATCH:
-	case TG_NEEDMORE:
-	    xerror(0, Ngt("%ls: unbound sequence"), keyseq);
-	    return Exit_FAILURE;
-	case TG_UNIQUE:
-	case TG_AMBIGUOUS:
-	    return print_binding_main(
-		    &le_modes[mode], keyseq, tg.value.cmdfunc);
-	default:
-	    assert(false);
+    if ((tg.type & TG_EXACTMATCH) && tg.matchlength == wcslen(keyseq)) {
+	return print_binding_main(
+		le_id_to_mode(mode), keyseq, tg.value.cmdfunc);
+    } else {
+	xerror(0, Ngt("%ls: unbound sequence"), keyseq);
+	return Exit_FAILURE;
     }
 }
 
@@ -506,12 +501,13 @@ int print_binding(le_mode_id_T mode, const wchar_t *keyseq)
 int print_binding_main(
 	void *mode, const wchar_t *keyseq, le_command_func_T *cmd)
 {
+    const char *format;
     char modechar;
     wchar_t *keyseqquote;
     const char *commandname;
     int r;
 
-    switch ((le_mode_id_T) ((le_mode_T *) mode - le_modes)) {
+    switch (le_mode_to_id(mode)) {
 	case LE_MODE_VI_INSERT:     modechar = 'v';  break;
 	case LE_MODE_VI_COMMAND:    modechar = 'a';  break;
 	case LE_MODE_VI_SEARCH:     modechar = 'V';  break;
@@ -520,9 +516,13 @@ int print_binding_main(
 	case LE_MODE_CHAR_EXPECT:   modechar = 'c';  break;
 	default:                    assert(false);
     }
+    if (keyseq[0] == L'-')
+	format = "bindkey -%c -- %ls %s\n";
+    else
+	format = "bindkey -%c %ls %s\n";
     keyseqquote = quote_sq(keyseq);
     commandname = get_command_name(cmd);
-    r = printf("bindkey -%c %ls %s\n", modechar, keyseqquote, commandname);
+    r = printf(format, modechar, keyseqquote, commandname);
     if (r < 0)
 	xerror(errno, Ngt("cannot print to standard output"));
     free(keyseqquote);
@@ -538,6 +538,7 @@ const char *get_command_name(le_command_func_T *command)
     return NULL;
 }
 
+#if YASH_ENABLE_HELP
 const char bindkey_help[] = Ngt(
 "bindkey - set or print key bindings for line-editing\n"
 "\tbindkey -aev [keyseq [command]]\n"
@@ -556,6 +557,7 @@ const char bindkey_help[] = Ngt(
 " -e --emacs\n"
 "\tSpecifies the \"emacs\" mode.\n"
 );
+#endif
 
 
 /* vim: set ts=8 sts=4 sw=4 noet tw=80: */
