@@ -58,35 +58,25 @@ enum le_state_T le_state;
 enum le_editstate_T le_editstate;
 
 
-/* Initializes line-editing.
- * Must be called before each call to `le_readline'.
- * Returns true iff successful, in which case `le_readline' must be called
- * afterward. */
-bool le_setup(void)
-{
-    le_keymap_init();
-    return isatty(STDIN_FILENO) && isatty(STDERR_FILENO)
-	&& le_setupterm(true)
-	&& le_set_terminal();
-}
-
-/* Prints the specified `prompt' and reads one line from the standard input.
- * This function can be called after and only after `le_setup' succeeded.
- * The `prompt' may contain backslash escapes specified in "input.c".
+/* Do line-editing using the specified prompts.
+ * This function must be called after and only after `le_setup' succeeded.
+ * The prompts may contain backslash escapes specified in "input.c".
  * The result is returned as a newly malloced wide string, including the
- * trailing newline. When EOF is encountered or on error, an empty string is
- * returned. NULL is returned when interrupted. */
-wchar_t *le_readline(const wchar_t *prompt)
+ * trailing newline. It is assigned to `*resultp' iff the return value is
+ * INPUT_OK. Returns INPUT_ERROR iff failed to set up the terminal. */
+inputresult_T le_readline(struct promptset_T prompt, wchar_t **resultp)
 {
-    wchar_t *resultline;
-
     assert(is_interactive_now);
     assert(le_state == LE_STATE_INACTIVE);
 
+    if (!isatty(STDIN_FILENO) || !isatty(STDERR_FILENO)
+	    || !le_setupterm(true) || !le_set_terminal())
+	return INPUT_ERROR;
+
     le_state = LE_STATE_ACTIVE;
+    le_keymap_init();
     le_editing_init();
     le_display_init(prompt);
-    le_display_maybe_promptsp();
     reader_init();
     le_editstate = LE_EDITSTATE_EDITING;
 
@@ -94,6 +84,7 @@ wchar_t *le_readline(const wchar_t *prompt)
 	read_next();
     while (le_editstate == LE_EDITSTATE_EDITING);
 
+    wchar_t *resultline;
     reader_finalize();
     le_display_finalize();
     resultline = le_editing_finalize();
@@ -101,19 +92,18 @@ wchar_t *le_readline(const wchar_t *prompt)
     le_state = LE_STATE_INACTIVE;
 
     switch (le_editstate) {
-	case LE_EDITSTATE_EDITING:
-	    assert(false);
 	case LE_EDITSTATE_DONE:
-	    break;
+	    *resultp = resultline;
+	    return INPUT_OK;
 	case LE_EDITSTATE_ERROR:
-	    resultline[0] = L'\0';
-	    break;
+	    free(resultline);
+	    return INPUT_EOF;
 	case LE_EDITSTATE_INTERRUPTED:
 	    free(resultline);
-	    resultline = NULL;
-	    break;
+	    return INPUT_INTERRUPTED;
+	default:
+	    assert(false);
     }
-    return resultline;
 }
 
 /* Clears the edit line and restores the terminal state.
@@ -135,9 +125,8 @@ void le_resume_readline(void)
 	le_state = LE_STATE_ACTIVE;
 	le_setupterm(true);
 	le_set_terminal();
-	le_display_maybe_promptsp();
 	le_display_update();
-	fflush(stderr);
+	le_display_flush();
     }
 }
 
@@ -149,7 +138,7 @@ void le_display_size_changed(void)
 	le_display_clear();
 	le_setupterm(true);
 	le_display_update();
-	fflush(stderr);
+	le_display_flush();
     }
 }
 
@@ -185,7 +174,8 @@ void reader_finalize(void)
 
 /* Reads the next byte from the standard input and take all the corresponding
  * actions.
- * May return without doing anything if a signal is caught, etc.
+ * May return without doing anything if a signal was caught, if the shell was
+ * interrupted, etc.
  * The caller must check `le_state' after this function returned. This function
  * should be called repeatedly while `le_state' is LE_STATE_EDITING. */
 void read_next(void)
@@ -200,9 +190,10 @@ void read_next(void)
     if (c)
 	goto direct_first_buffer;
 
-    /* wait for and read the next byte */
     le_display_update();
-    fflush(stderr);
+    le_display_flush();
+
+    /* wait for and read the next byte */
     if (keycode_ambiguous)
 	timeout = !wait_for_input(STDIN_FILENO, true, get_read_timeout());
     else
@@ -291,7 +282,7 @@ process_wide:
 		append_to_second_buffer(L'\0');
 		break;
 	    case (size_t) -1:  // conversion error
-		le_alert();
+		lebuf_print_alert(true);
 		memset(&reader_state, 0, sizeof reader_state);
 		sb_clear(&reader_first_buffer);
 		le_next_verbatim = false;
