@@ -24,6 +24,7 @@
 #include <errno.h>
 #include <stddef.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <term.h>
 #include <termios.h>
 #include <unistd.h>
@@ -33,6 +34,8 @@
 #include "../option.h"
 #include "../sig.h"
 #include "../util.h"
+#include "../variable.h"
+#include "display.h"
 #include "key.h"
 #include "terminfo.h"
 #include "trie.h"
@@ -44,7 +47,10 @@
 /* terminfo capabilities */
 #define TI_am      "am"
 #define TI_bel     "bel"
+#define TI_blink   "blink"
+#define TI_bold    "bold"
 #define TI_clear   "clear"
+#define TI_colors  "colors"
 #define TI_cols    "cols"
 #define TI_cr      "cr"
 #define TI_cub     "cub"
@@ -55,9 +61,11 @@
 #define TI_cuf1    "cuf1"
 #define TI_cuu     "cuu"
 #define TI_cuu1    "cuu1"
+#define TI_dim     "dim"
 #define TI_ed      "ed"
 #define TI_el      "el"
 #define TI_flash   "flash"
+#define TI_invis   "invis"
 #define TI_kBEG    "kBEG"
 #define TI_kCAN    "kCAN"
 #define TI_kCMD    "kCMD"
@@ -210,35 +218,43 @@
 #define TI_ktbc    "ktbc"
 #define TI_kund    "kund"
 #define TI_lines   "lines"
+#define TI_msgr    "msgr"
 #define TI_nel     "nel"
 #define TI_op      "op"
+#define TI_rev     "rev"
 #define TI_rmkx    "rmkx"
 #define TI_setab   "setab"
 #define TI_setaf   "setaf"
 #define TI_setb    "setb"
 #define TI_setf    "setf"
-#define TI_sgr     "sgr"
+#define TI_sgr0    "sgr0"
 #define TI_smkx    "smkx"
+#define TI_smso    "smso"
+#define TI_smul    "smul"
 #define TI_xenl    "xenl"
+#define TI_xmc     "xmc"
 
 
 /* This flag is set to true when the terminfo database needs to be refreshed
  * because the $TERM variable has been changed. */
 _Bool le_need_term_update = 1;
 
-/* Number of lines and columns in the current terminal. */
+/* Number of lines, columns and colors available in the current terminal. */
 /* Initialized in `le_setupterm'. */
-int le_lines, le_columns;
+int le_lines, le_columns, le_colors;
 
-/* Whether the terminal has the "am" and "xenl" flags set, respectively. */
-_Bool le_ti_am, le_ti_xenl;
+/* The value of the "xmc" capability. */
+int le_ti_xmc;
+
+/* Whether the terminal has the "xenl" and "msgr" flags set. */
+_Bool le_ti_xenl, le_ti_msgr;
 
 /* True if the meta key inputs character whose 8th bit is set. */
 /* Used only if the `shopt_le_convmeta' option is "auto". */
 _Bool le_meta_bit8;
 
 /* Strings sent by terminal when special key is pressed.
- * Values of entries are `keyseq'. */
+ * The values of entries are `keyseq'. */
 trie_T *le_keycodes = NULL;
 
 /* True if the terminal is set to the keyboard-transmit mode. */
@@ -248,12 +264,18 @@ static _Bool transmit_mode = 0;
 static inline int is_strcap_valid(const char *s)
     __attribute__((const));
 static void set_up_keycodes(void);
+static _Bool try_print_cap(const char *capname)
+    __attribute__((nonnull));
 static void move_cursor(char *capone, char *capmul, long count, int affcnt)
     __attribute__((nonnull));
 static _Bool move_cursor_1(char *capone, long count)
     __attribute__((nonnull));
 static _Bool move_cursor_mul(char *capmul, long count, int affcnt)
     __attribute__((nonnull));
+static void print_color_code(long color, char *seta, char *set)
+    __attribute__((nonnull));
+static void print_smkx(void);
+static void print_rmkx(void);
 static int putchar_stderr(int c);
 
 
@@ -281,8 +303,10 @@ _Bool le_setupterm(_Bool bypass)
 	struct winsize ws;
 	if (ioctl(STDERR_FILENO, TIOCGWINSZ, &ws) == 0
 		&& ws.ws_row > 0 && ws.ws_col > 0) {
-	    le_lines = ws.ws_row;
-	    le_columns = ws.ws_col;
+	    if (getenv(VAR_LINES) == NULL)
+		le_lines = ws.ws_row;
+	    if (getenv(VAR_COLUMNS) == NULL)
+		le_columns = ws.ws_col;
 	    return 1;
 	}
     }
@@ -290,33 +314,37 @@ _Bool le_setupterm(_Bool bypass)
     (void) bypass;
 #endif /* HAVE_TIOCGWINSZ */
 
-    le_need_term_update = 0;
     if (once)
 	del_curterm(cur_term);
     if (setupterm(NULL, STDERR_FILENO, &err) != OK)
 	return 0;
     once = 1;
 
+    if (tigetflag(TI_am) <= 0)                     return 0;
     if (!is_strcap_valid(tigetstr(TI_cub1))
-	    || !is_strcap_valid(tigetstr(TI_cub))) return 0;
+	    && !is_strcap_valid(tigetstr(TI_cub))) return 0;
     if (!is_strcap_valid(tigetstr(TI_cuf1))
-	    || !is_strcap_valid(tigetstr(TI_cuf))) return 0;
+	    && !is_strcap_valid(tigetstr(TI_cuf))) return 0;
     if (!is_strcap_valid(tigetstr(TI_cud1))
-	    || !is_strcap_valid(tigetstr(TI_cud))) return 0;
+	    && !is_strcap_valid(tigetstr(TI_cud))) return 0;
     if (!is_strcap_valid(tigetstr(TI_cuu1))
-	    || !is_strcap_valid(tigetstr(TI_cuu))) return 0;
-    if (!is_strcap_valid(tigetstr(TI_el))) return 0;
+	    && !is_strcap_valid(tigetstr(TI_cuu))) return 0;
+    if (!is_strcap_valid(tigetstr(TI_el)))         return 0;
 
     le_lines = tigetnum(TI_lines);
     le_columns = tigetnum(TI_cols);
-    le_ti_am = tigetflag(TI_am) > 0;
-    le_ti_xenl = tigetflag(TI_xenl) > 0;
-    le_meta_bit8 = tigetflag(TI_km) > 0;
     if (le_lines <= 0 || le_columns <= 0)
 	return 0;
 
+    le_colors = tigetnum(TI_colors);
+    le_ti_xmc = tigetnum(TI_xmc);
+    le_ti_xenl = tigetflag(TI_xenl) > 0;
+    le_ti_msgr = tigetflag(TI_msgr) > 0;
+    le_meta_bit8 = tigetflag(TI_km) > 0;
+
     set_up_keycodes();
 
+    le_need_term_update = 0;
     return 1;
 }
 
@@ -511,24 +539,43 @@ void set_up_keycodes(void)
     le_keycodes = t;
 }
 
-/* Prints the "cr" code. (carriage return: move cursor to first char of line) */
-void le_print_cr(void)
+/* Tries to print the specified capability string to the print buffer.
+ * Returns 1 iff successful. */
+_Bool try_print_cap(const char *capname)
 {
-    char *v = tigetstr(TI_cr);
-    if (is_strcap_valid(v))
-	tputs(v, 1, putchar_stderr);
-    else
-	fputc('\r', stderr);
+    char *v = tigetstr((char *) capname);
+    if (is_strcap_valid(v) && v[0] != '\0')
+	if (tputs(v, 1, lebuf_putchar) != ERR)
+	    return 1;
+    return 0;
 }
 
-/* Prints the "nel" code. (newline: move cursor to first char of next line) */
-void le_print_nel(void)
+/* Prints the "cr" code to the print buffer.
+ * (carriage return: move cursor to first char of line) */
+void lebuf_print_cr(void)
 {
-    char *v = tigetstr(TI_nel);
-    if (is_strcap_valid(v))
-	tputs(v, 1, putchar_stderr);
+#if 0
+    char *v = tigetstr(TI_cr);
+    if (is_strcap_valid(v) && v[0] != '\0')
+	tputs(v, 1, lebuf_putchar);
     else
-	fputc('\r', stderr), fputc('\n', stderr);
+#endif
+	lebuf_putchar('\r');
+    lebuf.pos.column = 0;
+}
+
+/* Prints the "nel" code to the print buffer.
+ * (newline: move cursor to first char of next line) */
+void lebuf_print_nel(void)
+{
+#if 0
+    char *v = tigetstr(TI_nel);
+    if (is_strcap_valid(v) && v[0] != '\0')
+	tputs(v, 1, lebuf_putchar);
+    else
+#endif
+	lebuf_putchar('\r'), lebuf_putchar('\n');
+    lebuf.pos.line++, lebuf.pos.column = 0;
 }
 
 /* Moves the cursor.
@@ -550,9 +597,9 @@ void move_cursor(char *capone, char *capmul, long count, int affcnt)
 _Bool move_cursor_1(char *capone, long count)
 {
     char *v = tigetstr(capone);
-    if (is_strcap_valid(v)) {
+    if (is_strcap_valid(v) && v[0] != '\0') {
 	do
-	    tputs(v, 1, putchar_stderr);
+	    tputs(v, 1, lebuf_putchar);
 	while (--count > 0);
 	return 1;
     } else {
@@ -566,121 +613,201 @@ _Bool move_cursor_mul(char *capmul, long count, int affcnt)
     if (is_strcap_valid(v)) {
 	v = tparm(v, count, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L);
 	if (v) {
-	    tputs(v, affcnt, putchar_stderr);
+	    tputs(v, affcnt, lebuf_putchar);
 	    return 1;
 	}
     }
     return 0;
 }
 
-/* Prints the "cub"/"cub1" code. (move cursor backward by `count' columns) */
-/* `count' must be small enough not to go beyond the screen bounds. */
-void le_print_cub(long count)
+/* Prints the "cub"/"cub1" code to the print buffer.
+ * (move cursor backward by `count' columns)
+ * `count' must be small enough not to go beyond the screen bounds. */
+void lebuf_print_cub(long count)
 {
     move_cursor(TI_cub1, TI_cub, count, 1);
+    lebuf.pos.column -= count;
+    assert(lebuf.pos.column >= 0);
 }
 
-/* Prints the "cuf"/"cuf1" code. (move cursor forward by `count' columns) */
-/* `count' must be small enough not to go beyond the screen bounds .*/
-void le_print_cuf(long count)
+/* Prints the "cuf"/"cuf1" code to the print buffer.
+ * (move cursor forward by `count' columns)
+ * `count' must be small enough not to go beyond the screen bounds. */
+void lebuf_print_cuf(long count)
 {
     move_cursor(TI_cuf1, TI_cuf, count, 1);
+    lebuf.pos.column += count;
+    assert(lebuf.pos.column < le_columns);
 }
 
-/* Prints the "cud"/"cud1" code. (move cursor down by `count' lines) */
-/* `count' must be small enough not to go beyond screen bounds .*/
-/* Note that this function may put the cursor at column 0 as side effect. */
-void le_print_cud(long count)
+/* Prints the "cud"/"cud1" code to the print buffer.
+ * (move cursor down by `count' lines)
+ * `count' must be small enough not to go beyond screen bounds.
+ * The cursor must be on the first column. */
+void lebuf_print_cud(long count)
 {
+    assert(lebuf.pos.column == 0);
     move_cursor(TI_cud1, TI_cud, count, count + 1);
+    lebuf.pos.line += count;
 }
 
-/* Prints the "cuu"/"cuu1" code. (move cursor up by `count' lines) */
-/* `count' must be small enough not to go beyond screen bounds .*/
-/* Note that this function may put the cursor at column 0 as side effect. */
-void le_print_cuu(long count)
+/* Prints the "cuu"/"cuu1" code to the print buffer.
+ * (move cursor up by `count' lines)
+ * `count' must be small enough not to go beyond screen bounds.
+ * The cursor must be on the first column. */
+void lebuf_print_cuu(long count)
 {
+    assert(lebuf.pos.column == 0);
     move_cursor(TI_cuu1, TI_cuu, count, count + 1);
+    lebuf.pos.line -= count;
 }
 
-/* Prints the "el" code. (clear to end of line) */
-void le_print_el(void)
-{
-    char *v = tigetstr(TI_el);
-    if (is_strcap_valid(v))
-	tputs(v, 1, putchar_stderr);
-}
-
-/* Prints the "ed" code if available. (clear to end of screen)
+/* Prints the "el" code to the print buffer. (clear to end of line)
  * Returns true iff successful. */
-_Bool le_print_ed(void)
+_Bool lebuf_print_el(void)
 {
-    char *v = tigetstr(TI_ed);
-    if (is_strcap_valid(v))
-	return tputs(v, 1, putchar_stderr) != ERR;
-    else
-	return 0;
+    return try_print_cap(TI_el);
+}
+
+/* Prints the "ed" code to the print buffer if available.
+ * (clear to end of screen)
+ * Returns true iff successful. */
+_Bool lebuf_print_ed(void)
+{
+    return try_print_cap(TI_ed);
 }
 
 /* Prints the "clear" code if available. (clear whole screen)
  * Returns true iff successful. */
-_Bool le_print_clear(void)
+_Bool lebuf_print_clear(void)
 {
-    char *v = tigetstr(TI_clear);
-    if (is_strcap_valid(v))
-	return tputs(v, le_lines, putchar_stderr) != ERR;
-    else
-	return 0;
+    return try_print_cap(TI_clear);
 }
 
-/* Prints the "sgr" code. Every argument must be 0 or 1. */
-void le_print_sgr(long standout, long underline, long reverse, long blink,
-	long dim, long bold, long invisible)
+/* Prints the "op" code. (set color pairs to default)
+ * Returns true iff successful. */
+_Bool lebuf_print_op(void)
 {
-    char *v = tigetstr(TI_sgr);
-    if (is_strcap_valid(v)) {
-	v = tparm(v, standout, underline, reverse, blink, dim, bold, invisible,
-		0L, 0L);
-	if (v)
-	    tputs(v, 1, putchar_stderr);
-    }
-}
-
-/* Prints the "op" code. (set color pairs to default) */
-void le_print_op(void)
-{
-    char *v = tigetstr(TI_op);
-    if (is_strcap_valid(v))
-	tputs(v, 1, putchar_stderr);
+    return try_print_cap(TI_op);
 }
 
 /* Prints the "setf"/"setaf" code. */
-void le_print_setfg(int color)
+void lebuf_print_setfg(long color)
 {
-    char *v = tigetstr(TI_setaf);
-    if (!is_strcap_valid(v))
-	v = tigetstr(TI_setf);
-    if (is_strcap_valid(v)) {
-	v = tparm(v, color, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L);
-	if (v)
-	    tputs(v, 1, putchar_stderr);
-    }
+    print_color_code(color, TI_setaf, TI_setf);
 }
 
 /* Prints the "setb"/"setab" code. */
-void le_print_setbg(int color)
+void lebuf_print_setbg(long color)
 {
-    char *v = tigetstr(TI_setab);
-    if (!is_strcap_valid(v))
-	v = tigetstr(TI_setb);
+    print_color_code(color, TI_setab, TI_setb);
+}
+
+void print_color_code(long color, char *seta, char *set)
+{
+    if (le_colors < 16)
+	color &= 0x7L;
+    if (le_colors <= color)
+	return;
+
+    char *v = tigetstr(seta);
+    if (!is_strcap_valid(v) || v[0] == '\0') {
+	v = tigetstr(set);
+	color = ((color & 0x1L) << 2)
+	      |  (color & 0x2L)
+	      | ((color & 0x4L) >> 2)
+	      |  (color & ~0x7L);
+    }
     if (is_strcap_valid(v)) {
 	v = tparm(v, color, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L);
 	if (v)
-	    tputs(v, 1, putchar_stderr);
+	    tputs(v, 1, lebuf_putchar);
     }
 }
 
-/* Prints the "smkx" code and sets the `transmit_mode' flag. */
+/* Prints the "sgr0" code to the print buffer.
+ * (reset the terminal font)
+ * Returns true iff successful. */
+_Bool lebuf_print_sgr0(void)
+{
+    return try_print_cap(TI_sgr0);
+}
+
+/* Prints the "smso" code to the print buffer. (start standout mode)
+ * Returns true iff successful. */
+_Bool lebuf_print_smso(void)
+{
+    if (try_print_cap(TI_smso)) {
+	if (le_ti_xmc > 0)
+	    lebuf_update_position(le_ti_xmc);
+	return 1;
+    } else {
+	return 0;
+    }
+}
+
+/* Prints the "smul" code to the print buffer. (start underline mode)
+ * Returns true iff successful. */
+_Bool lebuf_print_smul(void)
+{
+    return try_print_cap(TI_smul);
+}
+
+/* Prints the "rev" code to the print buffer. (start reverse mode)
+ * Returns true iff successful. */
+_Bool lebuf_print_rev(void)
+{
+    return try_print_cap(TI_rev);
+}
+
+/* Prints the "blink" code to the print buffer. (start blink mode)
+ * Returns true iff successful. */
+_Bool lebuf_print_blink(void)
+{
+    return try_print_cap(TI_blink);
+}
+
+/* Prints the "dim" code to the print buffer. (start dim mode)
+ * Returns true iff successful. */
+_Bool lebuf_print_dim(void)
+{
+    return try_print_cap(TI_dim);
+}
+
+/* Prints the "bold" code to the print buffer. (start bold mode)
+ * Returns true iff successful. */
+_Bool lebuf_print_bold(void)
+{
+    return try_print_cap(TI_bold);
+}
+
+/* Prints the "invis" code to the print buffer. (start invisible mode)
+ * Returns true iff successful. */
+_Bool lebuf_print_invis(void)
+{
+    return try_print_cap(TI_invis);
+}
+
+/* Prints the "flash" or "bel" code to alert the user.
+ * If `direct_stderr' is true, the output is sent to the standard error;
+ * otherwise, to the print buffer. */
+void lebuf_print_alert(_Bool direct_stderr)
+{
+    int (*outfunc)(int) = direct_stderr ? putchar_stderr : lebuf_putchar;
+    char *v = NULL;
+
+    if (shopt_le_visiblebell)
+	v = tigetstr(TI_flash);
+    if (!is_strcap_valid(v) || v[0] == '\0')
+	v = tigetstr(TI_bel);
+    if (is_strcap_valid(v) && v[0] == '\0')
+	tputs(v, 1, outfunc);
+    else
+	outfunc('\a');
+}
+
+/* Prints the "smkx" code to the standard error and sets the `transmit_mode'
+ * flag. */
 void print_smkx(void)
 {
     char *v = tigetstr(TI_smkx);
@@ -690,8 +817,8 @@ void print_smkx(void)
     }
 }
 
-/* Prints the "rmkx" code if the `transmit_mode' flag is set.
- * The flag is cleared in this function. */
+/* Prints the "rmkx" code to the standard error if the `transmit_mode' flag is
+ * set. The flag is cleared in this function. */
 void print_rmkx(void)
 {
     if (transmit_mode) {
@@ -708,21 +835,6 @@ int putchar_stderr(int c)
     return fputc(c, stderr);
 }
 
-/* Alerts the user by a flash or a bell without moving the cursor. */
-void le_alert(void)
-{
-    char *v = NULL;
-
-    if (shopt_le_visiblebell)
-	v = tigetstr(TI_flash);
-    if (!is_strcap_valid(v))
-	v = tigetstr(TI_bel);
-    if (is_strcap_valid(v))
-	tputs(v, 1, putchar_stderr);
-    else
-	putchar_stderr('\a');
-}
-
 
 /********** TERMIOS **********/
 
