@@ -22,28 +22,46 @@
 
 extern _Bool le_need_term_update;
 
-extern int le_lines, le_columns;
-extern _Bool le_ti_am, le_ti_xenl;
+extern int le_lines, le_columns, le_colors;
+extern int le_ti_xmc;
+extern _Bool le_ti_am, le_ti_xenl, le_ti_msgr;
 extern _Bool le_meta_bit8;
 extern struct trienode_T /* trie_T */ *le_keycodes;
 
 extern _Bool le_setupterm(_Bool bypass);
 
-extern void le_print_cr(void);
-extern void le_print_nel(void);
-extern void le_print_cub(long count);
-extern void le_print_cuf(long count);
-extern void le_print_cud(long count);
-extern void le_print_cuu(long count);
-extern void le_print_el(void);
-extern _Bool le_print_ed(void);
-extern _Bool le_print_clear(void);
-extern void le_print_sgr(long standout, long underline, long reverse,
-	long blink, long dim, long bold, long invisible);
-extern void le_print_op(void);
-extern void le_print_setfg(int color);
-extern void le_print_setbg(int color);
-extern void le_alert(void);
+enum le_color {
+    LE_COLOR_BLACK   = 0,
+    LE_COLOR_RED     = 1,
+    LE_COLOR_GREEN   = 2,
+    LE_COLOR_YELLOW  = 3,
+    LE_COLOR_BLUE    = 4,
+    LE_COLOR_MAGENTA = 5,
+    LE_COLOR_CYAN    = 6,
+    LE_COLOR_WHITE   = 7,
+};
+
+extern void lebuf_print_cr(void);
+extern void lebuf_print_nel(void);
+extern void lebuf_print_cub(long count);
+extern void lebuf_print_cuf(long count);
+extern void lebuf_print_cud(long count);
+extern void lebuf_print_cuu(long count);
+extern _Bool lebuf_print_el(void);
+extern _Bool lebuf_print_ed(void);
+extern _Bool lebuf_print_clear(void);
+extern _Bool lebuf_print_op(void);
+extern void lebuf_print_setfg(long color);
+extern void lebuf_print_setbg(long color);
+extern _Bool lebuf_print_sgr0(void);
+extern _Bool lebuf_print_smso(void);
+extern _Bool lebuf_print_smul(void);
+extern _Bool lebuf_print_rev(void);
+extern _Bool lebuf_print_blink(void);
+extern _Bool lebuf_print_dim(void);
+extern _Bool lebuf_print_bold(void);
+extern _Bool lebuf_print_invis(void);
+extern void lebuf_print_alert(_Bool direct_stderr);
 
 
 extern int le_eof_char, le_kill_char, le_interrupt_char, le_erase_char;
