@@ -55,9 +55,16 @@
 #include "yash.h"
 
 
-#if HAVE_EACCESS
+#if HAVE_FACCESSAT
+# ifndef faccessat
+extern int faccessat(int fd, const char *path, int amode, int flags)
+    __attribute__((nonnull));
+# endif
+#elif HAVE_EACCESS
+# ifndef eaccess
 extern int eaccess(const char *path, int amode)
     __attribute__((nonnull));
+# endif
 #else
 static bool check_access(const char *path, mode_t mode)
     __attribute__((nonnull));
@@ -90,7 +97,9 @@ bool is_irregular_file(const char *path)
 /* Checks if `path' is a readable file. */
 bool is_readable(const char *path)
 {
-#if HAVE_EACCESS
+#if HAVE_FACCESSAT
+    return faccessat(AT_FDCWD, path, R_OK, AT_EACCESS) == 0;
+#elif HAVE_EACCESS
     return eaccess(path, R_OK) == 0;
 #else
     return check_access(path, S_IRUSR | S_IRGRP | S_IROTH);
@@ -100,7 +109,9 @@ bool is_readable(const char *path)
 /* Checks if `path' is a writable file. */
 bool is_writable(const char *path)
 {
-#if HAVE_EACCESS
+#if HAVE_FACCESSAT
+    return faccessat(AT_FDCWD, path, W_OK, AT_EACCESS) == 0;
+#elif HAVE_EACCESS
     return eaccess(path, W_OK) == 0;
 #else
     return check_access(path, S_IWUSR | S_IWGRP | S_IWOTH);
@@ -110,14 +121,16 @@ bool is_writable(const char *path)
 /* Checks if `path' is an executable file (or a searchable directory). */
 bool is_executable(const char *path)
 {
-#if HAVE_EACCESS
+#if HAVE_FACCESSAT
+    return faccessat(AT_FDCWD, path, X_OK, AT_EACCESS) == 0;
+#elif HAVE_EACCESS
     return eaccess(path, X_OK) == 0;
 #else
     return check_access(path, S_IXUSR | S_IXGRP | S_IXOTH);
 #endif
 }
 
-#if !HAVE_EACCESS
+#if !HAVE_FACCESSAT && !HAVE_EACCESS
 /* Checks if this process has a proper permission to access the specified file.
  * Returns false if the file does not exist. */
 bool check_access(const char *path, mode_t mode)
@@ -161,7 +174,7 @@ bool check_access(const char *path, mode_t mode)
 
     return st.st_mode & S_IRWXO;
 }
-#endif /* !HAVE_EACCESS */
+#endif /* !HAVE_FACCESSAT && !HAVE_EACCESS */
 
 /* Checks if `path' is a readable regular file. */
 bool is_readable_regular(const char *path)
@@ -1091,15 +1104,12 @@ bool is_reentry(const struct stat *st, const struct wglob_dirstack *dirstack)
 
 /********** Builtins **********/
 
-static int change_directory(
-	const wchar_t *newpwd, bool printnewdir, bool logical)
-    __attribute__((nonnull,warn_unused_result));
 static void canonicalize_path_ex(xwcsbuf_T *buf)
     __attribute__((nonnull));
 static bool starts_with_root_parent(const wchar_t *path)
     __attribute__((nonnull,pure));
-static bool print_command_paths(bool all);
-static bool print_home_directories(void);
+static void print_command_paths(bool all);
+static void print_home_directories(void);
 static bool print_umask_octal(mode_t mode);
 static bool print_umask_symbolic(mode_t mode);
 static int set_umask(const wchar_t *newmask)
@@ -1111,64 +1121,65 @@ static inline mode_t copy_group_mask(mode_t mode)
 static inline mode_t copy_other_mask(mode_t mode)
     __attribute__((const));
 
-/* options for "cd", "pwd" and "pushd" builtins */
-static const struct xoption cd_pwd_options[] = {
-    { L"logical",  xno_argument, L'L', },
-    { L"physical", xno_argument, L'P', },
-#if YASH_ENABLE_HELP
-    { L"help",     xno_argument, L'-', },
-#endif
-    { NULL, 0, 0, },
-};
-
 /* The "cd" builtin, which accepts the following options:
  *  -L: don't resolve symlinks (default)
  *  -P: resolve symlinks
+ *  --default-directory=<dir>: go to <dir> when the operand is missing
  * -L and -P are mutually exclusive: the one specified last is used. */
 int cd_builtin(int argc, void **argv)
 {
     bool logical = true;
-    bool printnewdir = false;
-    const wchar_t *newpwd;
-    wchar_t opt;
+    const wchar_t *newpwd = NULL;
 
+    wchar_t opt;
     xoptind = 0, xopterr = true;
-    while ((opt = xgetopt_long(argv, L"-LP", cd_pwd_options, NULL))) {
+    while ((opt = xgetopt_long(argv, L"-LP", cd_options, NULL))) {
 	switch (opt) {
-	    case L'L':  logical = true;   break;
-	    case L'P':  logical = false;  break;
+	    case L'L':  logical = true;    break;
+	    case L'P':  logical = false;   break;
+	    case L'd':  newpwd = xoptarg;  break;
 #if YASH_ENABLE_HELP
 	    case L'-':
 		return print_builtin_help(ARGV(0));
 #endif
-	    default:
+	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  %ls [-L|-P] [dir]\n"), L"cd");
 		return Exit_ERROR;
 	}
     }
 
-    if (argc <= xoptind) {  /* step 1-2 */
-	newpwd = getvar(L VAR_HOME);
-	if (newpwd == NULL || newpwd[0] == L'\0') {
-	    xerror(0, Ngt("$HOME not set"));
-	    return Exit_FAILURE;
-	}
-    } else if (wcscmp(ARGV(xoptind), L"-") == 0) {
-	newpwd = getvar(L VAR_OLDPWD);
-	if (newpwd == NULL || newpwd[0] == L'\0') {
-	    xerror(0, Ngt("$OLDPWD not set"));
-	    return Exit_FAILURE;
-	}
-	printnewdir = true;
-    } else {
-	newpwd = ARGV(xoptind);
+    bool printnewdir = false;
+    switch (argc - xoptind) {
+	case 0:
+	    if (newpwd == NULL) {
+		newpwd = getvar(L VAR_HOME);
+		if (newpwd == NULL || newpwd[0] == L'\0') {
+		    xerror(0, Ngt("$HOME not set"));
+		    return Exit_FAILURE;
+		}
+	    }
+	    break;
+	case 1:
+	    if (wcscmp(ARGV(xoptind), L"-") == 0) {
+		newpwd = getvar(L VAR_OLDPWD);
+		if (newpwd == NULL || newpwd[0] == L'\0') {
+		    xerror(0, Ngt("$OLDPWD not set"));
+		    return Exit_FAILURE;
+		}
+		printnewdir = true;
+	    } else {
+		newpwd = ARGV(xoptind);
+	    }
+	    break;
+	default:
+	    goto print_usage;
     }
     return change_directory(newpwd, printnewdir, logical);
 }
 
 /* Changes the working directory to `newpwd'.
  * This function implements the almost whole part of the "cd" builtin.
- * $PWD, $OLDPWD and $DIRSTACK are set in this function.
+ * $PWD and $OLDPWD are set in this function.
  * If `printnewdir' is true or the new directory is found from $CDPATH, the new
  * directory is printed to stdout.
  * Returns Exit_SUCCESS, Exit_FAILURE or Exit_ERROR. */
@@ -1371,7 +1382,6 @@ const char cd_help[] = Ngt(
 "cd - change directory\n"
 "\tcd [-L|-P] [dir]\n"
 "Changes the working directory to <dir>.\n"
-"If <dir> is not specified, it defaults to $HOME.\n"
 "If <dir> is \"-\", the working directory is changed to $OLDPWD.\n"
 "If <dir> is a relative path which does not start with \".\" or \"..\",\n"
 "paths in $CDPATH are searched to find a new directory.\n"
@@ -1382,160 +1392,12 @@ const char cd_help[] = Ngt(
 "resolved so that $PWD does not contain any symbolic links.\n"
 "-L and -P are mutually exclusive: the last specified one is used.\n"
 "If neither is specified, -L is the default.\n"
+"If <dir> is not specified, the working directory is changed to the directory\n"
+"specified by the --default-directory=... option. If this option is not\n"
+"specified either, the default is $HOME.\n"
 );
 #endif
 
-#if YASH_ENABLE_DIRSTACK
-
-/* The "pushd" builtin.
- * The command syntax is the same as that of the "cd" builtin. */
-int pushd_builtin(int argc __attribute__((unused)), void **argv)
-{
-    bool logical = true;
-    bool printnewdir = false;
-    wchar_t opt;
-
-    xoptind = 0, xopterr = true;
-    while ((opt = xgetopt_long(argv, L"-LP", cd_pwd_options, NULL))) {
-	switch (opt) {
-	    case L'L':  logical = true;   break;
-	    case L'P':  logical = false;  break;
-#if YASH_ENABLE_HELP
-	    case L'-':
-		return print_builtin_help(ARGV(0));
-#endif
-	    default:
-		fprintf(stderr, gt("Usage:  %ls [-L|-P] [dir]\n"), L"pushd");
-		return Exit_ERROR;
-	}
-    }
-
-    const wchar_t *oldpwd = getvar(L VAR_PWD);
-    if (!oldpwd) {
-	xerror(0, Ngt("$PWD not set"));
-	return Exit_FAILURE;
-    }
-
-    const wchar_t *arg = ARGV(xoptind);
-    if (!arg)
-	arg = L"+1";
-
-    size_t stackindex;
-    const wchar_t *newpwd;
-    if (wcscmp(arg, L"-") == 0) {
-	printnewdir = true;
-	stackindex = SIZE_MAX;
-	newpwd = getvar(L VAR_OLDPWD);
-	if (!newpwd) {
-	    xerror(0, Ngt("$OLDPWD not set"));
-	    return Exit_FAILURE;
-	}
-    } else {
-	if (!parse_dirstack_index(arg, &stackindex, &newpwd, true))
-	    return Exit_FAILURE;
-    }
-    assert(newpwd != NULL);
-
-    wchar_t *saveoldpwd = xwcsdup(oldpwd);
-    int result = change_directory(newpwd, printnewdir, logical);
-#ifndef NDEBUG
-    newpwd = NULL;  /* newpwd cannot be used anymore. */
-#endif
-    if (result != Exit_SUCCESS) {
-	free(saveoldpwd);
-	return result;
-    }
-    if (!push_dirstack(saveoldpwd))
-	return Exit_FAILURE;
-    if (stackindex != SIZE_MAX)
-	if (!remove_dirstack_entry(stackindex))
-	    return Exit_FAILURE;
-    return Exit_SUCCESS;
-}
-
-#if YASH_ENABLE_HELP
-const char pushd_help[] = Ngt(
-"pushd - push directory into directory stack\n"
-"\tpushd [-L|-P] [dir]\n"
-"Changes the working directory to <dir> and appends it to the directory\n"
-"stack. The options and operand are basically the same as those of the \"cd\"\n"
-"builtin.\n"
-"If <dir> is an integer with the plus or minus sign, it is considered a\n"
-"specific entry of the stack, which is removed from the stack and appended\n"
-"again. An integer with the plus sign specifies the n'th newest entry, and\n"
-"a one with the minus sign specifies the n'th oldest entry.\n"
-"If <dir> is omitted, \"+1\" is assumed.\n"
-);
-#endif
-
-/* The "popd" builtin. */
-int popd_builtin(int argc __attribute__((unused)), void **argv)
-{
-    wchar_t opt;
-
-    xoptind = 0, xopterr = true;
-    while ((opt = xgetopt_long(argv, L"-", help_option, NULL))) {
-	switch (opt) {
-#if YASH_ENABLE_HELP
-	    case L'-':
-		return print_builtin_help(ARGV(0));
-#endif
-	    default:
-		fprintf(stderr, gt("Usage:  popd [index]\n"));
-		return Exit_ERROR;
-	}
-    }
-
-    const wchar_t *arg = ARGV(xoptind);
-    if (!arg)
-	arg = L"+0";
-
-    size_t stacksize = get_dirstack_size();
-    if (stacksize == 0) {
-	xerror(0, Ngt("directory stack is empty"));
-	return Exit_FAILURE;
-    }
-
-    size_t stackindex;
-    const wchar_t *dummy;
-    if (!parse_dirstack_index(arg, &stackindex, &dummy, true))
-	return Exit_FAILURE;
-    if (stackindex == SIZE_MAX) {
-	xerror(0, Ngt("`%ls' is not a valid index"), arg);
-	return Exit_ERROR;
-    }
-
-    int result;
-    if (stackindex == stacksize) {
-	wchar_t *newpwd = pop_dirstack();
-	if (!newpwd) {
-	    xerror(0, Ngt("cannot set directory stack"));
-	    return Exit_FAILURE;
-	}
-	result = change_directory(newpwd, true, true);
-	free(newpwd);
-	return result;
-    } else {
-	return remove_dirstack_entry(stackindex) ? Exit_SUCCESS : Exit_FAILURE;
-    }
-}
-
-#if YASH_ENABLE_HELP
-const char popd_help[] = Ngt(
-"popd - pop directory from directory stack\n"
-"\tpopd [index]\n"
-"Removes the last entry of the directory stack, returning to the previous\n"
-"directory.\n"
-"If <index> is an integer with the plus or minus sign, it is considered a\n"
-"specified entry of the stack, which is removed from the stack instead of the\n"
-"last entry. An integer with the plus sign specifies the n'th newest entry,\n"
-"and a one with the minus sign specifies the n'th oldest entry.\n"
-"If <index> is omitted, \"+0\" is assumed.\n"
-);
-#endif
-
-#endif /* YASH_ENABLE_DIRSTACK */
-
 /* The "pwd" builtin, which accepts the following options:
  *  -L: don't resolve symlinks (default)
  *  -P: resolve symlinks
@@ -1544,11 +1406,10 @@ int pwd_builtin(int argc __attribute__((unused)), void **argv)
 {
     bool logical = true;
     char *mbspwd;
-    int result;
     wchar_t opt;
 
     xoptind = 0, xopterr = true;
-    while ((opt = xgetopt_long(argv, L"-LP", cd_pwd_options, NULL))) {
+    while ((opt = xgetopt_long(argv, L"-LP", pwd_options, NULL))) {
 	switch (opt) {
 	    case L'L':  logical = true;   break;
 	    case L'P':  logical = false;  break;
@@ -1583,14 +1444,10 @@ int pwd_builtin(int argc __attribute__((unused)), void **argv)
 	return Exit_FAILURE;
     }
 print:
-    if (printf("%s\n", mbspwd) >= 0) {
-	result = Exit_SUCCESS;
-    } else {
+    if (printf("%s\n", mbspwd) < 0)
 	xerror(errno, Ngt("cannot print to standard output"));
-	result = Exit_FAILURE;
-    }
     free(mbspwd);
-    return result;
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 }
 
 #if YASH_ENABLE_HELP
@@ -1644,7 +1501,6 @@ int hash_builtin(int argc, void **argv)
     if (all && xoptind != argc)
 	goto print_usage;
 
-    bool err = false;
     if (dir) {
 	if (remove) {
 	    if (xoptind == argc) {  // forget all
@@ -1655,14 +1511,11 @@ int hash_builtin(int argc, void **argv)
 	    }
 	} else {
 	    if (xoptind == argc) {  // print all
-		err = print_home_directories();
+		print_home_directories();
 	    } else {                // remember the specified
-		for (int i = xoptind; i < argc; i++) {
-		    if (!get_home_directory(ARGV(i), true)) {
+		for (int i = xoptind; i < argc; i++)
+		    if (!get_home_directory(ARGV(i), true))
 			xerror(0, Ngt("%ls: no such user"), ARGV(i));
-			err = true;
-		    }
-		}
 	    }
 	}
     } else {
@@ -1680,29 +1533,26 @@ int hash_builtin(int argc, void **argv)
 	    }
 	} else {
 	    if (xoptind == argc) {  // print all
-		err = print_command_paths(all);
+		print_command_paths(all);
 	    } else {                // remember the specified
 		for (int i = xoptind; i < argc; i++) {
 		    if (wcschr(ARGV(i), L'/')) {
 			xerror(0, Ngt("%ls: command name must not contain `/'"),
 				ARGV(i));
-			err = true;
 			continue;
 		    }
 
 		    char *cmd = malloc_wcstombs(ARGV(i));
 		    if (cmd) {
-			if (!get_command_path(cmd, true)) {
+			if (!get_command_path(cmd, true))
 			    xerror(0, Ngt("%s: not found in $PATH"), cmd);
-			    err = true;
-			}
 			free(cmd);
 		    }
 		}
 	    }
 	}
     }
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 
 print_usage:
     fprintf(stderr, gt(posixly_correct
@@ -1713,8 +1563,9 @@ print_usage:
 }
 
 /* Prints the entries of the command hashtable.
- * Returns FALSE iff successful. */
-bool print_command_paths(bool all)
+ * Prints an error message to the standard error if failed to print to the
+ * standard output. */
+void print_command_paths(bool all)
 {
     kvpair_T kv;
     size_t index = 0;
@@ -1723,16 +1574,16 @@ bool print_command_paths(bool all)
 	if (all || !get_builtin(kv.key)) {
 	    if (printf("%s\n", (char *) kv.value) < 0) {
 		xerror(errno, Ngt("cannot print to standard output"));
-		return true;
+		break;
 	    }
 	}
     }
-    return false;
 }
 
 /* Prints the entries of the home directory hashtable.
- * Returns FALSE iff successful. */
-bool print_home_directories(void)
+ * Prints an error message to the standard error if failed to print to the
+ * standard output. */
+void print_home_directories(void)
 {
     kvpair_T kv;
     size_t index = 0;
@@ -1741,10 +1592,9 @@ bool print_home_directories(void)
 	const wchar_t *key = kv.key, *value = kv.value;
 	if (printf("~%ls=%ls\n", key, value) < 0) {
 	    xerror(errno, Ngt("cannot print to standard output"));
-	    return true;
+	    break;
 	}
     }
-    return false;
 }
 
 #if YASH_ENABLE_HELP
