@@ -81,7 +81,7 @@ static void reset_special_handler(
 	int signum, void (*handler)(int signum), bool leave);
 static void sig_handler(int signum);
 static void handle_sigchld(void);
-static bool set_trap(int signum, const wchar_t *command);
+static void set_trap(int signum, const wchar_t *command);
 static bool is_ignored(int signum);
 #if YASH_ENABLE_LINEEDIT && defined(SIGWINCH)
 static inline void handle_sigwinch(void);
@@ -97,33 +97,33 @@ bool process_exists(pid_t pid)
 
 /* Returns the name of the signal with the specified number.
  * The returned name doesn't have the "SIG"-prefix.
- * "?" is returned for an unknown signal number.
+ * An empty string is returned for an unknown signal number.
  * The returned string is valid until the next call to this function. */
-const char *get_signal_name(int signum)
+const wchar_t *get_signal_name(int signum)
 {
     if (signum == 0)
-	return "EXIT";
+	return L"EXIT";
 #if defined SIGRTMIN && defined SIGRTMAX
     int sigrtmin = SIGRTMIN, sigrtmax = SIGRTMAX;
     if (sigrtmin <= signum && signum <= sigrtmax) {
-	static char *name = NULL;
+	static wchar_t *name = NULL;
 	if (signum == sigrtmin)
-	    return "RTMIN";
+	    return L"RTMIN";
 	if (signum == sigrtmax)
-	    return "RTMAX";
+	    return L"RTMAX";
 	int range = sigrtmax - sigrtmin, diff = signum - sigrtmin;
 	free(name);
 	if (diff <= range / 2)
-	    name = malloc_printf("RTMIN+%d", diff);
+	    name = malloc_wprintf(L"RTMIN+%d", diff);
 	else
-	    name = malloc_printf("RTMAX-%d", sigrtmax - signum);
+	    name = malloc_wprintf(L"RTMAX-%d", sigrtmax - signum);
 	return name;
     }
 #endif
     for (const signal_T *s = signals; s->no; s++)
 	if (s->no == signum)
 	    return s->name;
-    return "?";
+    return L"";
 }
 
 /* Returns the number of the signal whose name is `name'.
@@ -132,11 +132,11 @@ const char *get_signal_name(int signum)
  * returned.
  * Returns 0 for "EXIT" and -1 for an unknown name.
  * `name' should be all in uppercase. */
-int get_signal_number(const char *name)
+int get_signal_number(const wchar_t *name)
 {
-    if (xisdigit(name[0])) {  /* parse a decimal integer */
+    if (iswdigit(name[0])) {  /* parse a decimal integer */
 	int signum;
-	if (!xstrtoi(name, 10, &signum) || signum < 0)
+	if (!xwcstoi(name, 10, &signum) || signum < 0)
 	    return -1;
 
 	/* check if `signum' is a valid signal */
@@ -152,37 +152,39 @@ int get_signal_number(const char *name)
 	return -1;
     }
 
-    if (strcmp(name, "EXIT") == 0)
+    if (wcscmp(name, L"EXIT") == 0)
 	return 0;
-    if (strncmp(name, "SIG", 3) == 0)
+    if (wcsncmp(name, L"SIG", 3) == 0)
 	name += 3;
     for (const signal_T *s = signals; s->no; s++)
-	if (strcmp(name, s->name) == 0)
+	if (wcscmp(name, s->name) == 0)
 	    return s->no;
 #if defined SIGRTMIN && defined SIGRTMAX
-    if (strncmp(name, "RTMIN", 5) == 0) {
+    if (wcsncmp(name, L"RTMIN", 5) == 0) {
 	int sigrtmin = SIGRTMIN;
 	name += 5;
-	if (name[0] == '\0') {
+	if (name[0] == L'\0') {
 	    return sigrtmin;
-	} else if (name[0] == '+') {
+	} else if (name[0] == L'+') {
 	    int num;
-	    if (!xstrtoi(name, 10, &num)
-		    || num < 0 || SIGRTMAX - sigrtmin < num)
-		return -1;
-	    return sigrtmin + num;
+	    if (xwcstoi(name, 10, &num)) {
+		assert(num >= 0);
+		if (SIGRTMAX - sigrtmin >= num)
+		    return sigrtmin + num;
+	    }
 	}
-    } else if (strncmp(name, "RTMAX", 5) == 0) {
+    } else if (wcsncmp(name, L"RTMAX", 5) == 0) {
 	int sigrtmax = SIGRTMAX;
 	name += 5;
-	if (name[0] == '\0') {
+	if (name[0] == L'\0') {
 	    return sigrtmax;
-	} else if (name[0] == '-') {
+	} else if (name[0] == L'-') {
 	    int num;
-	    if (!xstrtoi(name, 10, &num)
-		    || num > 0 || SIGRTMIN - sigrtmax > num)
-		return -1;
-	    return sigrtmax + num;
+	    if (xwcstoi(name, 10, &num)) {
+		assert(num < 0);
+		if (SIGRTMIN - sigrtmax <= num)
+		    return sigrtmax + num;
+	    }
 	}
     }
 #endif
@@ -190,18 +192,16 @@ int get_signal_number(const char *name)
 }
 
 /* Returns the number of the signal whose name is `name'.
+ * `name' may have the "SIG"-prefix.
+ * If `name' is an integer that is a valid signal number, the number is
+ * returned.
  * Returns 0 for "EXIT" and -1 for an unknown name.
  * The given string is converted into uppercase. */
-int get_signal_number_w(wchar_t *name)
+int get_signal_number_toupper(wchar_t *name)
 {
     for (wchar_t *n = name; *n; n++)
 	*n = towupper(*n);
-    char *mname = malloc_wcstombs(name);
-    if (!mname)
-	return -1;
-    int result = get_signal_number(mname);
-    free(mname);
-    return result;
+    return get_signal_number(name);
 }
 
 
@@ -712,13 +712,13 @@ void execute_exit_trap(void)
  * If `command' is NULL, the trap is reset to the default.
  * If `command' is an empty string, the trap is set to SIG_IGN.
  * This function may call `get_signal_name'.
- * Returns true iff successful. */
-bool set_trap(int signum, const wchar_t *command)
+ * An error message is printed to the standard error on error. */
+void set_trap(int signum, const wchar_t *command)
 {
     if (signum == SIGKILL || signum == SIGSTOP) {
-	xerror(0, Ngt("SIG%s: cannot be trapped"),
-		signum == SIGKILL ? "KILL" : "STOP");
-	return false;
+	xerror(0, Ngt("SIG%ls: cannot be trapped"),
+		signum == SIGKILL ? L"KILL" : L"STOP");
+	return;
     }
 
     wchar_t **commandp;
@@ -731,9 +731,9 @@ bool set_trap(int signum, const wchar_t *command)
 	    commandp = &rttrap_command[index];
 	    receivedp = &rtsignal_received[index];
 	} else {
-	    xerror(0, Ngt("SIG%s: unsupported real-time signal"),
+	    xerror(0, Ngt("SIG%ls: unsupported real-time signal"),
 		    get_signal_name(signum));
-	    return false;
+	    return;
 	}
     } else
 #endif
@@ -747,12 +747,9 @@ bool set_trap(int signum, const wchar_t *command)
 	/* Signals that were ignored on entry to a non-interactive shell cannot
 	 * be trapped or reset. (POSIX) */
 #if FIXED_SIGNAL_AS_ERROR
-	xerror(0, Ngt("SIG%s: cannot be reset"),
-		get_signal_name(signum));
-	return false;
-#else
-	return true;
+	xerror(0, Ngt("SIG%ls: cannot be reset"), get_signal_name(signum));
 #endif
+	return;
     }
 
     /* If `*commandp' is currently executed, we must not free it. */
@@ -767,7 +764,7 @@ bool set_trap(int signum, const wchar_t *command)
     }
     *receivedp = false;
     if (signum == 0)
-	return true;
+	return;
 
     struct sigaction action;
     if (command == NULL)
@@ -793,7 +790,7 @@ bool set_trap(int signum, const wchar_t *command)
     switch (signum) {
 	case SIGCHLD:
 	    /* SIGCHLD's signal handler is always `sig_handler' */
-	    return true;
+	    return;
 	case SIGINT:
 #if YASH_ENABLE_LINEEDIT && defined(SIGWINCH)
 	case SIGWINCH:
@@ -801,7 +798,7 @@ bool set_trap(int signum, const wchar_t *command)
 	    /* SIGINT and SIGWINCH's signal handler is always `sig_handler'
 	     * when interactive */
 	    if (interactive_handlers_set)
-		return true;
+		return;
 	    break;
 	case SIGTSTP:
 	    if (job_handler_set)
@@ -823,12 +820,9 @@ default_ignore:
 
     sigemptyset(&action.sa_mask);
     action.sa_flags = 0;
-    if (sigaction(signum, &action, NULL) >= 0) {
-	return true;
-    } else {
+    if (sigaction(signum, &action, NULL) < 0) {
 	int saveerrno = errno;
-	xerror(saveerrno, "sigaction(SIG%s)", get_signal_name(signum));
-	return false;
+	xerror(saveerrno, "sigaction(SIG%ls)", get_signal_name(signum));
     }
 }
 
@@ -930,10 +924,11 @@ void reset_sigwinch(void)
 
 /********** Builtin **********/
 
-static bool print_trap(const char *signame, const wchar_t *command)
+static bool print_trap(const wchar_t *signame, const wchar_t *command)
     __attribute__((nonnull(1)));
-static bool print_signal(int signum, const char *name, bool verbose);
-static bool signal_job(int signum, const wchar_t *jobname)
+static bool print_signal(int signum, const wchar_t *name, bool verbose)
+    __attribute__((nonnull));
+static void signal_job(int signum, const wchar_t *jobname)
     __attribute__((nonnull));
 
 /* The "trap" builtin */
@@ -971,7 +966,7 @@ int trap_builtin(int argc, void **argv)
 	/* print all traps */
 	sigset_t ss;
 	sigemptyset(&ss);
-	if (!print_trap("EXIT", trap_command[sigindex(0)]))
+	if (!print_trap(L"EXIT", trap_command[sigindex(0)]))
 	    return Exit_FAILURE;
 	for (const signal_T *s = signals; s->no; s++) {
 	    if (!sigismember(&ss, s->no)) {
@@ -992,39 +987,33 @@ int trap_builtin(int argc, void **argv)
 	return Exit_SUCCESS;
     } else if (print) {
 	/* print specified traps */
-	bool ok = true;
 #if defined SIGRTMIN && defined SIGRTMAX
 	int sigrtmin = SIGRTMIN, sigrtmax = SIGRTMAX;
 #endif
+	clearerr(stdout);
 	do {
-	    wchar_t *wname = ARGV(xoptind);
-	    for (wchar_t *w = wname; *w; w++)
-		*w = towupper(*w);
-	    char *name = malloc_wcstombs(wname);
-	    if (!name)
-		name = "";
-	    int signum = get_signal_number(name);
+	    wchar_t *name = ARGV(xoptind);
+	    int signum = get_signal_number_toupper(name);
 	    if (signum < 0) {
-		xerror(0, Ngt("%ls: no such signal"), wname);
+		xerror(0, Ngt("%ls: no such signal"), name);
 	    } else {
 #if defined SIGRTMIN && defined SIGRTMAX
 		if (sigrtmin <= signum && signum <= sigrtmax) {
 		    int index = signum - sigrtmin;
 		    if (index < RTSIZE)
-			ok = print_trap(get_signal_name(signum),
+			print_trap(get_signal_name(signum),
 				rttrap_command[index]);
 		} else
 #endif
 		{
-		    ok = print_trap(name, trap_command[sigindex(signum)]);
+		    print_trap(name, trap_command[sigindex(signum)]);
 		}
 	    }
-	    free(name);
-	} while (ok && ++xoptind < argc);
-	return ok ? Exit_SUCCESS : Exit_FAILURE;
+	} while (!ferror(stdout) && ++xoptind < argc);
+
+	return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
     }
 
-    bool err = false;
     const wchar_t *command;
 
     /* check if the first operand is an integer */
@@ -1045,14 +1034,13 @@ int trap_builtin(int argc, void **argv)
 set_traps:
     do {
 	wchar_t *name = ARGV(xoptind);
-	int signum = get_signal_number_w(name);
-	if (signum < 0) {
+	int signum = get_signal_number_toupper(name);
+	if (signum >= 0)
+	    set_trap(signum, command);
+	else
 	    xerror(0, Ngt("%ls: no such signal"), name);
-	} else {
-	    err |= !set_trap(signum, command);
-	}
     } while (++xoptind < argc);
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 
 print_usage:
     if (posixly_correct)
@@ -1072,12 +1060,13 @@ print_usage:
  * the current signal handler for the specified signal.
  * If the `command' is NULL, this function does nothing.
  * Otherwise, the `command' is properly single-quoted and printed.
- * Returns true iff successful (no error). */
-bool print_trap(const char *signame, const wchar_t *command)
+ * Returns true iff successful (no error). On error, an error message is printed
+ * to the standard error. */
+bool print_trap(const wchar_t *signame, const wchar_t *command)
 {
-    if (command) {
+    if (command != NULL) {
 	wchar_t *q = quote_sq(command);
-	int r = printf("trap -- %ls %s\n", q, signame);
+	int r = printf("trap -- %ls %ls\n", q, signame);
 	if (r < 0)
 	    xerror(errno, Ngt("cannot print to standard output"));
 	free(q);
@@ -1122,7 +1111,6 @@ int kill_builtin(int argc, void **argv)
     wchar_t opt;
     int signum = SIGTERM;
     bool list = false, verbose = false;
-    bool err = false;
 
     xoptind = 0, xopterr = false;
     while ((opt = xgetopt_long(argv,
@@ -1138,7 +1126,7 @@ int kill_builtin(int argc, void **argv)
 				"without `SIG'"), xoptarg);
 		    return Exit_ERROR;
 		}
-		signum = get_signal_number_w(xoptarg);
+		signum = get_signal_number_toupper(xoptarg);
 		if (signum < 0 || (signum == 0 && !iswdigit(xoptarg[0]))) {
 		    xerror(0, Ngt("%ls: no such signal"), xoptarg);
 		    return Exit_FAILURE;
@@ -1167,17 +1155,21 @@ int kill_builtin(int argc, void **argv)
 
 main:
     if (list) {
-	/* print signal info */
 	if (xoptind == argc) {
+	    /* print info of all signals */
 	    for (const signal_T *s = signals; s->no; s++)
-		print_signal(s->no, s->name, verbose);
+		if (!print_signal(s->no, s->name, verbose))
+		    return Exit_FAILURE;
 #if defined SIGRTMIN && defined SIGRTMAX
 	    for (int i = SIGRTMIN, max = SIGRTMAX; i <= max; i++)
-		print_signal(i, NULL, verbose);
+		if (!print_signal(i, get_signal_name(i), verbose))
+		    return Exit_FAILURE;
 #endif
 	} else {
+	    /* print info of the specified signals */
 	    do {
 		int signum;
+		const wchar_t *signame;
 
 		if (xwcstoi(ARGV(xoptind), 10, &signum) && signum >= 0) {
 		    if (signum >= TERMSIGOFFSET)
@@ -1185,12 +1177,13 @@ main:
 		    else if (signum >= (TERMSIGOFFSET & 0xFF))
 			signum -= (TERMSIGOFFSET & 0xFF);
 		} else {
-		    signum = get_signal_number_w(ARGV(xoptind));
+		    signum = get_signal_number_toupper(ARGV(xoptind));
 		}
-		if (signum <= 0 || !print_signal(signum, NULL, verbose)) {
+		signame = get_signal_name(signum);
+		if (signum <= 0 || signame[0] == L'\0')
 		    xerror(0, Ngt("%ls: no such signal"), ARGV(xoptind));
-		    err = true;
-		}
+		else if (!print_signal(signum, signame, verbose))
+		    return Exit_FAILURE;
 	    } while (++xoptind < argc);
 	}
     } else {
@@ -1200,26 +1193,23 @@ main:
 	do {
 	    wchar_t *proc = ARGV(xoptind);
 	    if (proc[0] == L'%') {
-		if (!signal_job(signum, proc))
-		    err = true;
+		signal_job(signum, proc);
 	    } else {
 		long pid;
 
 		if (!xwcstol(proc, 10, &pid)) {
 		    xerror(0, Ngt("`%ls' is not a valid integer"), proc);
-		    err = true;
 		    continue;
 		}
 		// XXX this cast might not be safe
 		if (kill((pid_t) pid, signum) < 0) {
 		    xerror(errno, "%ls", proc);
-		    err = true;
 		    continue;
 		}
 	    }
 	} while (++xoptind < argc);
     }
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 
 print_usage:
     if (posixly_correct)
@@ -1234,46 +1224,42 @@ print_usage:
 }
 
 /* Prints info about the specified signal.
- * If `name' is non-NULL, it must be the correct name of signal `signum'.
- * If `name' is NULL and `signum' is not a valid signal number, it is an error.
- * `get_signal_name' may be called in this function.
- * Returns true iff successful. */
-bool print_signal(int signum, const char *name, bool verbose)
+ * `signum' must be a valid signal number and `name' must be the name of the
+ * signal.
+ * Returns false iff an IO error occurred. */
+bool print_signal(int signum, const wchar_t *name, bool verbose)
 {
-    if (!name) {
-	name = get_signal_name(signum);
-	if (strcmp(name, "?") == 0)
-	    return false;
-    }
+    int r;
+
     if (!verbose) {
-	puts(name);
+	r = printf("%ls\n", name);
     } else {
 #if HAVE_STRSIGNAL
 	const char *sigdesc = strsignal(signum);
-	if (sigdesc) {
-	    printf("%d\t%-10s %s\n", signum, name, sigdesc);
-	} else
+	if (sigdesc)
+	    r = printf("%d\t%-10ls %s\n", signum, name, sigdesc);
+	else
 #endif
-	{
-	    printf("%d\t%-10s\n", signum, name);
-	}
+	    r = printf("%d\t%-10ls\n", signum, name);
+    }
+    if (r >= 0) {
+	return true;
+    } else {
+	xerror(errno, Ngt("cannot print to standard output"));
+	return false;
     }
-    return true;
 }
 
 /* Sends the specified signal to the specified job.
- * Returns true iff successful. */
-bool signal_job(int signum, const wchar_t *jobspec)
+ * Returns true iff successful. On error, an error message is printed to the
+ * standard error. */
+void signal_job(int signum, const wchar_t *jobspec)
 {
     pid_t jobpgid = get_job_pgid(jobspec);
     if (jobpgid <= 0)
-	return false;
-
-    if (kill(-jobpgid, signum) < 0) {
+	return;
+    if (kill(-jobpgid, signum) < 0)
 	xerror(errno, "%ls", jobspec);
-	return false;
-    }
-    return true;
 }
 
 #if YASH_ENABLE_HELP
