@@ -33,12 +33,12 @@
 
 
 /* If the type of the return value of the functions below is string buffer,
- * the return value is the argument `buf'. */
+ * the return value is the argument buffer. */
 
 
 /********** Multibyte String Buffer **********/
 
-/* Initializes a multibyte string buffer as a new empty string. */
+/* Initializes the specified string buffer as an empty string. */
 xstrbuf_T *sb_init(xstrbuf_T *buf)
 {
     buf->contents = xmalloc((XSTRBUF_INITSIZE + 1) * sizeof (char));
@@ -48,9 +48,10 @@ xstrbuf_T *sb_init(xstrbuf_T *buf)
     return buf;
 }
 
-/* Initializes a multibyte string buffer with an already malloced string.
- * After calling this function, the string is used as the buffer
- * so you must not touch or `free' it any more. */
+/* Initializes the specified multibyte string buffer with the specified string.
+ * String `s' must be `free'able.
+ * After calling this function, the string is used as the buffer, so you must
+ * not touch or `free' it any more. */
 xstrbuf_T *sb_initwith(xstrbuf_T *restrict buf, char *restrict s)
 {
     buf->contents = s;
@@ -58,18 +59,9 @@ xstrbuf_T *sb_initwith(xstrbuf_T *restrict buf, char *restrict s)
     return buf;
 }
 
-/* Frees a multibyte string buffer and returns the contents.
- * The caller must `free' the return value. */
-char *sb_tostr(xstrbuf_T *buf)
-{
-    if (buf->maxlength - buf->length > 20)
-	sb_setmax(buf, buf->length);
-    return buf->contents;
-}
-
-/* Changes `buf->maxlength'.
+/* Changes the max length of the specified buffer.
  * If `newmax' is less than the current length of the buffer, the end of
- * the buffered string is truncated. */
+ * the buffer contents is truncated. */
 xstrbuf_T *sb_setmax(xstrbuf_T *buf, size_t newmax)
 {
     buf->contents = xrealloc(buf->contents, (newmax + 1) * sizeof (char));
@@ -82,30 +74,23 @@ xstrbuf_T *sb_setmax(xstrbuf_T *buf, size_t newmax)
 
 /* If `buf->maxlength' is less than `max', reallocates the buffer so that
  * `buf->maxlength' is no less than `max'. */
-inline xstrbuf_T *sb_ensuremax(xstrbuf_T *buf, size_t max)
+xstrbuf_T *sb_ensuremax(xstrbuf_T *buf, size_t max)
 {
-    if (buf->maxlength < max) {
-	size_t newmax = buf->maxlength;
-	do
-	    newmax = newmax * 2 + 1;
-	while (newmax < max);
-	return sb_setmax(buf, newmax);
-    } else {
+    if (max <= buf->maxlength)
 	return buf;
-    }
+
+    size_t len15 = buf->maxlength + (buf->maxlength >> 1);
+    if (max < len15)
+	max = len15;
+    if (max < buf->maxlength + 10)
+	max = buf->maxlength + 10;
+    return sb_setmax(buf, max);
 }
 
-/* Clears the contents of a string buffer, preserving its `maxlength'. */
-xstrbuf_T *sb_clear(xstrbuf_T *buf)
-{
-    buf->contents[buf->length = 0] = '\0';
-    return buf;
-}
-
-/* Replaces the contents of a string buffer with another string.
- * `bn' characters starting at the offset `i' in the buffer is removed and
+/* Replaces the specified part of the buffer with another string.
+ * `bn' characters starting at offset `i' in buffer `buf' is removed and
  * the first `sn' characters of `s' take place of them.
- * No boundary checks are done and a null character is not considered special.
+ * No boundary checks are done and null characters are not considered special.
  * `s' must not be part of `buf->contents'. */
 xstrbuf_T *sb_replace_force(
 	xstrbuf_T *restrict buf, size_t i, size_t bn,
@@ -120,12 +105,12 @@ xstrbuf_T *sb_replace_force(
     return buf;
 }
 
-/* Replaces the contents of a string buffer with another string.
- * `bn' characters starting at the offset `i' in the buffer is removed and
+/* Replaces the specified part of the buffer with another string.
+ * `bn' characters starting at offset `i' in buffer `buf' is removed and
  * the first `sn' characters of `s' take place of them.
  * If (strlen(s) < sn), the whole of `s' is replaced with.
- * If (buf->length < i + sn), all the characters after the offset `i' in the
- * buffer is replaced. Especially, if (buf->length <= i), `s' is appended.
+ * If (buf->length < i + sn), all the characters after offset `i' in the buffer
+ * is replaced. Especially, if (buf->length <= i), `s' is appended.
  * `s' must not be part of `buf->contents'. */
 xstrbuf_T *sb_replace(
 	xstrbuf_T *restrict buf, size_t i, size_t bn,
@@ -139,8 +124,8 @@ xstrbuf_T *sb_replace(
     return sb_replace_force(buf, i, bn, s, sn);
 }
 
-/* Appends a byte `c' to the end of a string buffer.
- * For `c', a null character is not treated specially. */
+/* Appends byte `c' to the end of string buffer `buf'.
+ * The byte is appended even if it is a null byte. */
 xstrbuf_T *sb_ccat(xstrbuf_T *buf, char c)
 {
     sb_ensuremax(buf, buf->length + 1);
@@ -149,8 +134,8 @@ xstrbuf_T *sb_ccat(xstrbuf_T *buf, char c)
     return buf;
 }
 
-/* Appends `n' bytes of `c' to the end of a string buffer.
- * For `c', a null character is not treated specially. */
+/* Appends `n' bytes of `c' to the end of buffer `buf'.
+ * The bytes are appended even if `c' is a null byte. */
 xstrbuf_T *sb_ccat_repeat(xstrbuf_T *buf, char c, size_t n)
 {
     sb_ensuremax(buf, buf->length + n);
@@ -160,10 +145,11 @@ xstrbuf_T *sb_ccat_repeat(xstrbuf_T *buf, char c, size_t n)
     return buf;
 }
 
-/* Converts a wide character to a multibyte character and appends it to the
- * buffer. The shift state `*ps' is updated to describe the new state.
+/* Converts wide character `c' into a multibyte string and appends it to buffer
+ * `buf'. Shift state `ps' is used for the conversion.
  * Returns true iff successful.
- * If `c' is a null character, the shift state is reset to the initial state. */
+ * If `c' is a null character, the shift state is reset to the initial state but
+ * the null character is not appended to the buffer. */
 bool sb_wccat(xstrbuf_T *restrict buf, wchar_t c, mbstate_t *restrict ps)
 {
     size_t count;
@@ -183,53 +169,41 @@ bool sb_wccat(xstrbuf_T *restrict buf, wchar_t c, mbstate_t *restrict ps)
     return true;
 }
 
-/* Converts a wide string to a multibyte string and appends it to the buffer.
- * The shift state `*ps' is updated to describe the new state.
- * Returns NULL if the whole string is converted and appended successfully, or,
- * on error, a pointer to the character in `s' that caused the error.
- * A partial result may have been appended to the buffer on error.
- * If `s' is NULL, the shift state is restored to an initial state. */
+/* Converts wide string `s' to a multibyte string and appends it to buffer
+ * `buf'. Shift state `ps' is used for the conversion. After successful
+ * conversion, `ps' will be the initial shift state.
+ * Returns NULL if the whole string is converted and appended successfully,
+ * otherwise a pointer to the character in `s' that caused the error.
+ * A partial result may be left in the buffer on error. */
 wchar_t *sb_wcscat(xstrbuf_T *restrict buf,
 	const wchar_t *restrict s, mbstate_t *restrict ps)
 {
-    size_t count;
-
-    if (s) {
-	for (;;) {
-	    count = wcsrtombs(buf->contents + buf->length,
-		    (const wchar_t **) &s,
-		    buf->maxlength - buf->length + 1,
-		    ps);
-	    if (count == (size_t) -1)
-		break;
-	    buf->length += count;
-	    if (!s)
-		break;
-	    sb_ensuremax(buf, buf->maxlength * 2);
-	}
-    } else {
-	mbstate_t saveps = *ps;
-	count = wcrtomb(NULL, L'\0', &saveps);
+    for (;;) {
+	size_t count = wcsrtombs(buf->contents + buf->length,
+		(const wchar_t **) &s,
+		buf->maxlength - buf->length + 1,
+		ps);
 	if (count == (size_t) -1) {
 	    buf->contents[buf->length] = '\0';
-	    return NULL;
+	    break;
 	}
-	sb_ensuremax(buf, buf->length + count - 1);
-	count = wcrtomb(buf->contents + buf->length, L'\0', ps);
-	assert(0 < count && count - 1 <= buf->maxlength - buf->length);
-	buf->length += count - 1;
+	buf->length += count;
+	if (s == NULL)
+	    break;
+	sb_ensuremax(buf, buf->maxlength + 1);
     }
-
     assert(buf->contents[buf->length] == '\0');
     return (wchar_t *) s;
 }
 
-/* Appends a formatted string to the end of a buffer.
- * `format' and the following arguments must not be part of `buf->contents'. */
+/* Appends the result of `vsprintf' to the specified buffer.
+ * `format' and the following arguments must not be part of `buf->contents'.
+ * Returns the number of appended bytes if successful.
+ * On error, the buffer is not changed and -1 is returned. */
 int sb_vprintf(xstrbuf_T *restrict buf, const char *restrict format, va_list ap)
 {
-    va_list saveap;
-    va_copy(saveap, ap);
+    va_list copyap;
+    va_copy(copyap, ap);
 
     int rest = buf->maxlength - buf->length + 1;
     int result = vsnprintf(buf->contents + buf->length, rest, format, ap);
@@ -238,18 +212,22 @@ int sb_vprintf(xstrbuf_T *restrict buf, const char *restrict format, va_list ap)
 	/* If the buffer is too small... */
 	sb_ensuremax(buf, buf->length + result);
 	rest = buf->maxlength - buf->length + 1;
-	result = vsnprintf(buf->contents + buf->length, rest, format, saveap);
+	result = vsnprintf(buf->contents + buf->length, rest, format, copyap);
     }
     assert(result < rest);
     if (result >= 0)
 	buf->length += result;
+    else
+	buf->contents[buf->length] = '\0';
     assert(buf->contents[buf->length] == '\0');
-    va_end(saveap);
+    va_end(copyap);
     return result;
 }
 
-/* Appends a formatted string to the end of a buffer.
- * `format' and any following arguments must not be part of `buf->contents'. */
+/* Appends the result of `sprintf' to the specified buffer.
+ * `format' and the following arguments must not be part of `buf->contents'.
+ * Returns the number of appended bytes if successful.
+ * On error, the buffer is not changed and -1 is returned. */
 int sb_printf(xstrbuf_T *restrict buf, const char *restrict format, ...)
 {
     va_list ap;
@@ -261,37 +239,10 @@ int sb_printf(xstrbuf_T *restrict buf, const char *restrict format, ...)
     return result;
 }
 
-/* Appends the result of `strftime' to the end of a string buffer.
- * The result of formatting must not be an empty string.
- * Returns the number of the appended characters. */
-size_t sb_strftime(xstrbuf_T *restrict buf,
-	const char *restrict format, const struct tm *restrict tm)
-{
-    /* try with a 40-byte-long buffer. */
-    char result[40];
-    size_t count = strftime(result, sizeof result, format, tm);
-    if (count) {
-	sb_cat(buf, result);
-	return count;
-    }
-
-    /* if 40 bytes are too small, retry with a longer buffer. */
-    size_t len = sizeof result;
-    do {
-	len *= 2;
-	sb_ensuremax(buf, buf->length + len);
-	count = strftime(buf->contents + buf->length,
-		buf->maxlength - buf->length + 1,
-		format,
-		tm);
-    } while (count == 0);
-    return count;
-}
-
 
 /********** Wide String Buffer **********/
 
-/* Initializes a wide string buffer as a new empty string. */
+/* Initializes the specified wide string buffer as an empty string. */
 xwcsbuf_T *wb_init(xwcsbuf_T *buf)
 {
     buf->contents = xmalloc((XWCSBUF_INITSIZE + 1) * sizeof (wchar_t));
@@ -301,9 +252,10 @@ xwcsbuf_T *wb_init(xwcsbuf_T *buf)
     return buf;
 }
 
-/* Initializes a wide string buffer with an already malloced string.
- * After calling this function, the string is used as the buffer
- * so you must not touch or `free' it any more. */
+/* Initializes the specified wide string buffer with the specified string.
+ * String `s' must be `free'able.
+ * After calling this function, the string is used as the buffer, so you must
+ * not touch or `free' it any more. */
 xwcsbuf_T *wb_initwith(xwcsbuf_T *restrict buf, wchar_t *restrict s)
 {
     buf->contents = s;
@@ -311,18 +263,9 @@ xwcsbuf_T *wb_initwith(xwcsbuf_T *restrict buf, wchar_t *restrict s)
     return buf;
 }
 
-/* Frees a wide string buffer and returns the contents.
- * The caller must `free' the return value. */
-wchar_t *wb_towcs(xwcsbuf_T *buf)
-{
-    if (buf->maxlength - buf->length > 20)
-	wb_setmax(buf, buf->length);
-    return buf->contents;
-}
-
-/* Changes `buf->maxlength'.
+/* Changes the max length of the specified buffer.
  * If `newmax' is less than the current length of the buffer, the end of
- * the buffered string is truncated. */
+ * the buffer contents is truncated. */
 xwcsbuf_T *wb_setmax(xwcsbuf_T *buf, size_t newmax)
 {
     buf->contents = xrealloc(buf->contents, (newmax + 1) * sizeof (wchar_t));
@@ -335,30 +278,23 @@ xwcsbuf_T *wb_setmax(xwcsbuf_T *buf, size_t newmax)
 
 /* If `buf->maxlength' is less than `max', reallocates the buffer so that
  * `buf->maxlength' is no less than `max'. */
-inline xwcsbuf_T *wb_ensuremax(xwcsbuf_T *buf, size_t max)
+xwcsbuf_T *wb_ensuremax(xwcsbuf_T *buf, size_t max)
 {
-    if (buf->maxlength < max) {
-	size_t newmax = buf->maxlength;
-	do
-	    newmax = newmax * 2 + 1;
-	while (newmax < max);
-	return wb_setmax(buf, newmax);
-    } else {
+    if (max <= buf->maxlength)
 	return buf;
-    }
+
+    size_t len15 = buf->maxlength + (buf->maxlength >> 1);
+    if (max < len15)
+	max = len15;
+    if (max < buf->maxlength + 8)
+	max = buf->maxlength + 8;
+    return wb_setmax(buf, max);
 }
 
-/* Clears the contents of a string buffer, preserving its `maxlength'. */
-xwcsbuf_T *wb_clear(xwcsbuf_T *buf)
-{
-    buf->contents[buf->length = 0] = L'\0';
-    return buf;
-}
-
-/* Replaces the contents of a string buffer with another string.
- * `bn' characters starting at the offset `i' in the buffer is removed and
+/* Replaces the specified part of the buffer with another string.
+ * `bn' characters starting at offset `i' in buffer `buf' is removed and
  * the first `sn' characters of `s' take place of them.
- * No boundary checks are done and a null character is not considered special.
+ * No boundary checks are done and null characters are not considered special.
  * `s' must not be part of `buf->contents'. */
 xwcsbuf_T *wb_replace_force(
 	xwcsbuf_T *restrict buf, size_t i, size_t bn,
@@ -373,12 +309,12 @@ xwcsbuf_T *wb_replace_force(
     return buf;
 }
 
-/* Replaces the contents of a string buffer with another string.
- * `bn' characters starting at the offset `i' in the buffer is removed and
+/* Replaces the specified part of the buffer with another string.
+ * `bn' characters starting at offset `i' in buffer `buf' is removed and
  * the first `sn' characters of `s' take place of them.
  * If (wcslen(s) < sn), the whole of `s' is replaced with.
- * If (buf->length < i + sn), all the characters after the offset `i' in the
- * buffer is replaced. Especially, if (buf->length <= i), `s' is appended.
+ * If (buf->length < i + sn), all the characters after offset `i' in the buffer
+ * is replaced. Especially, if (buf->length <= i), `s' is appended.
  * `s' must not be part of `buf->contents'. */
 xwcsbuf_T *wb_replace(
 	xwcsbuf_T *restrict buf, size_t i, size_t bn,
@@ -392,8 +328,8 @@ xwcsbuf_T *wb_replace(
     return wb_replace_force(buf, i, bn, s, sn);
 }
 
-/* Appends a wide character `c' to the end of a string buffer.
- * For `c', a null character is not treated specially. */
+/* Appends wide character `c' to the end of buffer `buf'.
+ * The character is appended even if it is a null wide character. */
 xwcsbuf_T *wb_wccat(xwcsbuf_T *buf, wchar_t c)
 {
     wb_ensuremax(buf, buf->length + 1);
@@ -402,8 +338,8 @@ xwcsbuf_T *wb_wccat(xwcsbuf_T *buf, wchar_t c)
     return buf;
 }
 
-/* Appends `n' characters of `c' to the end of a string buffer.
- * For `c', a null character is not treated specially. */
+/* Appends `n' characters of `c' to the end of buffer `buf'.
+ * The characters are appended even if `c' is a null wide character. */
 xwcsbuf_T *wb_wccat_repeat(xwcsbuf_T *buf, wchar_t c, size_t n)
 {
     wb_ensuremax(buf, buf->length + n);
@@ -413,17 +349,17 @@ xwcsbuf_T *wb_wccat_repeat(xwcsbuf_T *buf, wchar_t c, size_t n)
     return buf;
 }
 
-/* Converts a multibyte string to a wide string and append it to the buffer.
- * The multibyte string is assumed to start in a initial shift state.
- * Returns NULL if the whole string is converted and appended successfully, or,
- * on error, a pointer to the character in `s' that caused the error.
- * A partial result may have been appended to the buffer on error. */
+/* Converts multibyte string `s' into a wide string and appends it to buffer
+ * `buf'. The multibyte string is assumed to start in the initial shift state.
+ * Returns NULL if the whole string is converted and appended successfully,
+ * otherwise a pointer to the character in `s' that caused the error.
+ * A partial result may be left in the buffer on error. */
 char *wb_mbscat(xwcsbuf_T *restrict buf, const char *restrict s)
 {
     mbstate_t state;
     size_t count;
 
-    memset(&state, 0, sizeof state);  /* initialize as a initial shift state */
+    memset(&state, 0, sizeof state);  // initialize as the initial shift state
 
     for (;;) {
 	count = mbsrtowcs(buf->contents + buf->length, (const char **) &s,
@@ -431,48 +367,51 @@ char *wb_mbscat(xwcsbuf_T *restrict buf, const char *restrict s)
 	if (count == (size_t) -1)
 	    break;
 	buf->length += count;
-	if (!s)
+	if (s == NULL)
 	    break;
-	wb_ensuremax(buf, buf->maxlength * 2);
+	wb_ensuremax(buf, buf->maxlength + 1);
     }
 
     buf->contents[buf->length] = L'\0';
     return (char *) s;
 }
 
-/* Appends a formatted string to the end of a buffer.
- * `format' and the following arguments must not be part of `buf->contents'. */
+/* Appends the result of `vswprintf' to the specified buffer.
+ * `format' and the following arguments must not be part of `buf->contents'.
+ * Returns the number of appended characters if successful.
+ * On error, the buffer is not changed and -1 is returned. */
 int wb_vwprintf(
 	xwcsbuf_T *restrict buf, const wchar_t *restrict format, va_list ap)
 {
-    va_list saveap;
-    va_copy(saveap, ap);
-
+    va_list copyap;
     int rest, result;
 
     for (int i = 0; i < 10; i++) {
+	va_copy(copyap, ap);
 	rest = buf->maxlength - buf->length + 1;
 	result = vswprintf(buf->contents + buf->length, rest, format, ap);
+	va_end(copyap);
 
 	if (0 <= result && result < rest)
 	    break;
 
 	/* According to POSIX, if the buffer is too short, `vswprintf' returns
-	 * a negative integer. However, on some systems, it returns a desired
-	 * buffer length like `vsprintf', which is rather preferable. */
+	 * a negative integer. On some systems, however, it returns a desired
+	 * buffer length as `vsprintf' does, which is rather preferable. */
 	wb_ensuremax(buf, buf->length + (result < 0 ? 2 * rest : result));
-	va_end(ap);
-	va_copy(ap, saveap);
     }
     if (result >= 0)
 	buf->length += result;
+    else
+	buf->contents[buf->length] = L'\0';
     assert(buf->contents[buf->length] == L'\0');
-    va_end(saveap);
     return result;
 }
 
-/* Appends a formatted string to the end of a buffer.
- * `format' and the following arguments must not be part of `buf->contents'. */
+/* Appends the result of `swprintf' to the specified buffer.
+ * `format' and the following arguments must not be part of `buf->contents'.
+ * Returns the number of appended characters if successful.
+ * On error, the buffer is not changed and -1 is returned. */
 int wb_wprintf(xwcsbuf_T *restrict buf, const wchar_t *restrict format, ...)
 {
     va_list ap;
@@ -488,27 +427,33 @@ int wb_wprintf(xwcsbuf_T *restrict buf, const wchar_t *restrict format, ...)
 
 /********** Multibyte-Wide Conversion Utilities **********/
 
-/* Converts a wide string to a newly malloced multibyte string.
+/* Converts the specified wide string into a newly malloced multibyte string.
  * Only the first `n' characters of `s' is converted at most.
  * Returns NULL on error.
  * The resulting string starts and ends in the initial shift state.*/
 char *malloc_wcsntombs(const wchar_t *s, size_t n)
+{
+    size_t nn = xwcsnlen(s, n);
+    if (s[nn] == L'\0')
+	return malloc_wcstombs(s);
+
+    wchar_t ss[nn + 1];
+    wcsncpy(ss, s, nn);
+    ss[nn] = L'\0';
+    return malloc_wcstombs(ss);
+}
+
+/* Converts the specified wide string into a newly malloced multibyte string.
+ * Returns NULL on error.
+ * The resulting string starts and ends in the initial shift state.*/
+char *malloc_wcstombs(const wchar_t *s)
 {
     xstrbuf_T buf;
     mbstate_t state;
-    size_t nn = xwcsnlen(s, n);
-    wchar_t ss[nn + 1];
-
-    if (s[nn] != L'\0') {
-	wcsncpy(ss, s, nn);
-	ss[nn] = L'\0';
-	s = ss;
-    }
 
     sb_init(&buf);
-    memset(&state, 0, sizeof state);  /* initialize as a initial shift state */
-    if (sb_wcscat(&buf, s, &state) == NULL
-	    && sb_wcscat(&buf, NULL, &state) == NULL) {
+    memset(&state, 0, sizeof state);  // initialize as the initial shift state
+    if (sb_wcscat(&buf, s, &state) == NULL) {
 	return sb_tostr(&buf);
     } else {
 	sb_destroy(&buf);
@@ -516,21 +461,27 @@ char *malloc_wcsntombs(const wchar_t *s, size_t n)
     }
 }
 
-/* Converts a multibyte string to a newly malloced wide string.
+/* Converts the specified multibyte string into a newly malloced wide string.
  * Only the first `n' bytes of `s' is converted at most.
  * Returns NULL on error.
- * The resulting string starts and ends in the initial shift state.*/
+ * The multibyte string is assumed to start in the initial shift state. */
 wchar_t *malloc_mbsntowcs(const char *s, size_t n)
 {
-    xwcsbuf_T buf;
     size_t nn = xstrnlen(s, n);
-    char ss[nn + 1];
+    if (s[nn] == '\0')
+	return malloc_mbstowcs(s);
 
-    if (s[nn] != '\0') {
-	strncpy(ss, s, nn);
-	ss[nn] = '\0';
-	s = ss;
-    }
+    char ss[nn + 1];
+    strncpy(ss, s, nn);
+    ss[nn] = '\0';
+    return malloc_mbstowcs(ss);
+}
+
+/* Converts the specified multibyte string into a newly malloced wide string.
+ * Returns NULL on error. */
+wchar_t *malloc_mbstowcs(const char *s)
+{
+    xwcsbuf_T buf;
 
     wb_init(&buf);
     if (wb_mbscat(&buf, s) == NULL) {
@@ -542,4 +493,29 @@ wchar_t *malloc_mbsntowcs(const char *s, size_t n)
 }
 
 
+/********** Formatting Utilities **********/
+
+/* Returns the result of `sprintf' as a newly malloced string. */
+char *malloc_printf(const char *format, ...)
+{
+    va_list ap;
+    char *result;
+    va_start(ap, format);
+    result = malloc_vprintf(format, ap);
+    va_end(ap);
+    return result;
+}
+
+/* Returns the result of `swprintf' as a newly malloced string. */
+wchar_t *malloc_wprintf(const wchar_t *format, ...)
+{
+    va_list ap;
+    wchar_t *result;
+    va_start(ap, format);
+    result = malloc_vwprintf(format, ap);
+    va_end(ap);
+    return result;
+}
+
+
 /* vim: set ts=8 sts=4 sw=4 noet tw=80: */
