@@ -17,12 +17,12 @@
 
 # make temporary directory
 : ${TMPDIR:=$PWD}
-if [ x"${TMPDIR}" = x"${TMPDIR#/}" ]; then
-    echo \$TMPDIR must be an absolute path >&2
-    exit 1
-fi
+case "$TMPDIR" in
+    /*) ;;
+    *)  echo "\$TMPDIR must be an absolute path" >&2; exit 1;;
+esac
 TESTTMP="$(cd ${TMPDIR}; pwd)/test.$$"
-trap 'rm -rf $TESTTMP; exit' EXIT HUP INT QUIT ABRT ALRM TERM PIPE USR1 USR2
+trap 'rm -rf "$TESTTMP"; exit' EXIT HUP INT QUIT ABRT ALRM TERM PIPE USR1 USR2
 if ! mkdir -m u=rwx,go= "$TESTTMP"; then
     echo Cannot create temporary directory >&2
     trap - EXIT
@@ -84,10 +84,10 @@ failed=0
 for x in $TEST_ITEMS
 do
     x="${x%.tst}"
-    if [ x"$x" = x"${x%.p}" ]
-    then INVOKE=
-    else INVOKE='./invoke sh'
-    fi
+    case "$x" in
+	*.p) INVOKE='./invoke sh' ;;
+	*  ) INVOKE=              ;;
+    esac
     if ! checkskip "$x"
     then
 	echo " * $x (skipped)"
