@@ -43,6 +43,7 @@
 #include "parser.h"
 #include "path.h"
 #include "plist.h"
+#include "sig.h"
 #include "strbuf.h"
 #include "util.h"
 #include "variable.h"
@@ -284,8 +285,6 @@ void init_variables(void)
 	    !posixly_correct ? L"\\$ " : (geteuid() != 0) ? L"$ " : L"# ";
 	set_variable(L VAR_PS1, xwcsdup(ps1),   SCOPE_GLOBAL, false);
 	set_variable(L VAR_PS2, xwcsdup(L"> "), SCOPE_GLOBAL, false);
-	if (!posixly_correct)
-	    set_variable(L VAR_PS3, xwcsdup(L"#? "), SCOPE_GLOBAL, false);
 	set_variable(L VAR_PS4, xwcsdup(L"+ "), SCOPE_GLOBAL, false);
     }
 
@@ -561,7 +560,8 @@ variable_T *new_temporary(const wchar_t *name)
 /* Creates a new variable with the specified name if there is none.
  * If the variable already exists, it is cleared and returned.
  *
- * On error, NULL is returned. Otherwise a (new) variable is returned.
+ * On error, an error message is printed to the standard error and NULL is
+ * returned. Otherwise, a (new) variable is returned.
  * `v_type' is the only valid member of the variable and the all members
  * (including `v_type') must be initialized by the caller. If `v_type' of the
  * return value has the VF_EXPORT flag set, the caller must call
@@ -591,7 +591,8 @@ variable_T *new_variable(const wchar_t *name, scope_T scope)
  * If `export' is true, the variable is exported (i.e., the VF_EXPORT flag is
  * set to the variable). But this function does not reset an existing VF_EXPORT
  * flag if `export' is false.
- * Returns true iff successful. */
+ * Returns true iff successful. On error, an error message is printed to the
+ * standard error. */
 bool set_variable(
 	const wchar_t *name, wchar_t *value, scope_T scope, bool export)
 {
@@ -620,13 +621,15 @@ bool set_variable(
  * `values' and its elements must be `free'able.
  * `count' is the number of elements in `values'. If `count' is zero, the
  * number is counted in this function.
- * Returns true iff successful. */
-bool set_array(const wchar_t *name, size_t count, void **values, scope_T scope)
+ * Returns the set array iff successful. On error, an error message is printed
+ * to the standard error and NULL is returned. */
+variable_T *set_array(
+	const wchar_t *name, size_t count, void **values, scope_T scope)
 {
     variable_T *var = new_variable(name, scope);
     if (!var) {
 	recfree(values, free);
-	return false;
+	return NULL;
     }
 
     var->v_type = VF_ARRAY | (var->v_type & (VF_EXPORT | VF_NODELETE));
@@ -637,7 +640,7 @@ bool set_array(const wchar_t *name, size_t count, void **values, scope_T scope)
     variable_set(name, var);
     if (var->v_type & VF_EXPORT)
 	update_environment(name);
-    return true;
+    return var;
 }
 
 /* Changes the value of the specified array element.
@@ -990,10 +993,18 @@ void variable_set(const wchar_t *name, variable_T *var)
     case L'C':
 	if (wcscmp(name, L VAR_CDPATH) == 0)
 	    reset_path(PA_CDPATH, var);
+#if YASH_ENABLE_LINEEDIT
+	if (wcscmp(name, L VAR_COLUMNS) == 0)
+	    le_need_term_update = true;
+#endif
 	break;
     case L'L':
 	if (wcscmp(name, L VAR_LANG) == 0 || wcsncmp(name, L"LC_", 3) == 0)
 	    reset_locale(name);
+#if YASH_ENABLE_LINEEDIT
+	if (wcscmp(name, L VAR_LINES) == 0)
+	    le_need_term_update = true;
+#endif
 	break;
     case L'M':
 	if (wcscmp(name, L VAR_MAILPATH) == 0)
@@ -1246,46 +1257,6 @@ void tryhash_word_as_command(const wordunit_T *w)
 
 #if YASH_ENABLE_DIRSTACK
 
-static variable_T *get_dirstack(void);
-
-/* Returns the directory stack.
- * If the stack is not yet created, it is initialized as an empty stack and
- * returned.
- * Fails if a non-array variable or a readonly array already exists, in which
- * case NULL is returned. */
-variable_T *get_dirstack(void)
-{
-    variable_T *var = search_variable(L VAR_DIRSTACK);
-    if (var) {
-	if (((var->v_type & VF_MASK) == VF_ARRAY)
-		&& !(var->v_type & VF_READONLY))
-	    return var;
-	else
-	    return NULL;
-    }
-
-    void **ary = xmalloc(1 * sizeof *ary);
-    ary[0] = NULL;
-    if (set_array(L VAR_DIRSTACK, 0, ary, SCOPE_GLOBAL)) {
-	var = search_variable(L VAR_DIRSTACK);
-	assert(var != NULL);
-	assert((var->v_type & VF_MASK) == VF_ARRAY);
-	assert(!(var->v_type & VF_READONLY));
-	return var;
-    } else {
-	return NULL;
-    }
-}
-
-/* Returns the number of elements in $DIRSTACK. */
-size_t get_dirstack_size(void)
-{
-    variable_T *var = search_variable(L VAR_DIRSTACK);
-    if (!var || ((var->v_type & VF_MASK) != VF_ARRAY))
-	return 0;
-    return var->v_valc;
-}
-
 /* Parses the specified string as the index of a directory stack entry.
  * The index string must start with the plus or minus sign.
  * If the corresponding entry is found, the entry is assigned to `*entryp' and
@@ -1297,8 +1268,9 @@ size_t get_dirstack_size(void)
  * Returns true if successful (`*entryp' and `*indexp' are assigned).
  * Returns false on error, in which case an error message is printed iff
  * `printerror' is true. */
-bool parse_dirstack_index(const wchar_t *indexstr,
-	size_t *indexp, const wchar_t **entryp, bool printerror)
+bool parse_dirstack_index(
+	const wchar_t *restrict indexstr, size_t *restrict indexp,
+	const wchar_t **restrict entryp, bool printerror)
 {
     long num;
 
@@ -1366,86 +1338,29 @@ out_of_range:
     return false;
 }
 
-/* Adds the specified value to the directory stack ($DIRSTACK).
- * A new stack is created if there is none.
- * `value' is freed in this function.
- * Returns true iff successful. */
-bool push_dirstack(wchar_t *value)
-{
-    variable_T *var = get_dirstack();
-    if (!var) {
-	free(value);
-	xerror(0, Ngt("cannot set directory stack"));
-	return false;
-    }
-
-    size_t index = var->v_valc++;
-    var->v_vals = xrealloc(var->v_vals, (index + 2) * sizeof *var->v_vals);
-    var->v_vals[index] = value;
-    var->v_vals[index + 1] = NULL;
-    return true;
-}
-
-/* Removes the newest entry of the directory stack ($DIRSTACK) and returns it.
- * The return value must be freed by the caller.
- * Returns NULL on error. */
-wchar_t *pop_dirstack(void)
-{
-    variable_T *var = search_variable(L VAR_DIRSTACK);
-    if (!var || ((var->v_type & VF_MASK) != VF_ARRAY)
-	    || (var->v_type & VF_READONLY) || (var->v_valc == 0))
-	return NULL;
-
-    wchar_t *result = var->v_vals[--var->v_valc];
-    var->v_vals[var->v_valc] = NULL;
-    return result;
-}
-
-/* Removes the specified entry of the directory stack.
- * Returns true iff the entry was successfully removed or no such entry was
- * found. */
-bool remove_dirstack_entry(size_t index)
-{
-    variable_T *var = search_variable(L VAR_DIRSTACK);
-    if (!var || ((var->v_type & VF_MASK) != VF_ARRAY))
-	return true;
-    if (var->v_type & VF_READONLY) {
-	xerror(0, Ngt("cannot set directory stack"));
-	return false;
-    }
-
-    if (index < var->v_valc) {
-	free(var->v_vals[index]);
-	memmove(var->v_vals + index, var->v_vals + index + 1,
-		(var->v_valc - index) * sizeof *var->v_vals);
-	var->v_valc--;
-    }
-    return true;
-}
-
 #endif /* YASH_ENABLE_DIRSTACK */
 
 
 /********** Builtins **********/
 
-static bool print_variable(
+static void print_variable(
 	const wchar_t *name, const variable_T *var,
 	const wchar_t *argv0, bool readonly, bool export)
     __attribute__((nonnull));
-static bool print_function(
+static void print_function(
 	const wchar_t *name, const function_T *func,
 	const wchar_t *argv0, bool readonly)
     __attribute__((nonnull));
 #if YASH_ENABLE_ARRAY
-static bool array_remove_elements(
+static void array_remove_elements(
 	variable_T *array, size_t count, void *const *indexwcss)
     __attribute__((nonnull));
 static int compare_long(const void *lp1, const void *lp2)
     __attribute__((nonnull,pure));
-static bool array_insert_elements(
+static void array_insert_elements(
 	variable_T *array, size_t count, void *const *values)
     __attribute__((nonnull));
-static bool array_set_element(const wchar_t *name, variable_T *array,
+static void array_set_element(const wchar_t *name, variable_T *array,
 	const wchar_t *indexword, const wchar_t *value)
     __attribute__((nonnull));
 #endif /* YASH_ENABLE_ARRAY */
@@ -1461,7 +1376,7 @@ static bool set_to(const wchar_t *varname, wchar_t value)
     __attribute__((nonnull));
 static bool read_input(xwcsbuf_T *buf, bool noescape)
     __attribute__((nonnull));
-static bool split_and_assign_array(const wchar_t *name, wchar_t *values,
+static void split_and_assign_array(const wchar_t *name, wchar_t *values,
 	const wchar_t *ifs, bool raw)
     __attribute__((nonnull));
 
@@ -1540,7 +1455,6 @@ int typeset_builtin(int argc, void **argv)
 	return Exit_ERROR;
     }
 
-    bool ok = true;
     if (xoptind == argc) {
 	kvpair_T *kvs;
 	size_t count;
@@ -1555,17 +1469,16 @@ int typeset_builtin(int argc, void **argv)
 	    count = table.count;
 	    ht_destroy(&table);
 	    qsort(kvs, count, sizeof *kvs, keywcscoll);
-	    for (size_t i = 0; ok && i < count; i++)
-		ok &= print_variable(
+	    for (size_t i = 0; yash_error_message_count == 0 && i < count; i++)
+		print_variable(
 			kvs[i].key, kvs[i].value, ARGV(0), readonly, export);
 	} else {
 	    /* print all functions */
 	    kvs = ht_tokvarray(&functions);
 	    count = functions.count;
 	    qsort(kvs, count, sizeof *kvs, keywcscoll);
-	    for (size_t i = 0; ok && i < count; i++)
-		ok &= print_function(
-			kvs[i].key, kvs[i].value, ARGV(0), readonly);
+	    for (size_t i = 0; yash_error_message_count == 0 && i < count; i++)
+		print_function(kvs[i].key, kvs[i].value, ARGV(0), readonly);
 	}
 	free(kvs);
     } else {
@@ -1578,7 +1491,6 @@ int typeset_builtin(int argc, void **argv)
 		/* treat function */
 		if (wequal) {
 		    xerror(0, Ngt("cannot assign function"));
-		    ok = false;
 		    continue;
 		}
 
@@ -1587,11 +1499,9 @@ int typeset_builtin(int argc, void **argv)
 		    if (readonly)
 			f->f_type |= VF_READONLY | VF_NODELETE;
 		    if (print)
-			ok &= print_function(arg, f, ARGV(0), readonly);
+			print_function(arg, f, ARGV(0), readonly);
 		} else {
 		    xerror(0, Ngt("%ls: no such function"), arg);
-		    ok = false;
-		    continue;
 		}
 	    } else if (wequal || !print) {
 		/* create/assign variable */
@@ -1600,7 +1510,6 @@ int typeset_builtin(int argc, void **argv)
 		if (wequal) {
 		    if (var->v_type & VF_READONLY) {
 			xerror(0, Ngt("%ls: readonly"), arg);
-			ok = false;
 		    } else {
 			varvaluefree(var);
 			var->v_type = VF_NORMAL | (var->v_type & ~VF_MASK);
@@ -1622,37 +1531,34 @@ int typeset_builtin(int argc, void **argv)
 		/* print the variable */
 		variable_T *var = search_variable(arg);
 		if (var) {
-		    ok &= print_variable(arg, var, ARGV(0), readonly, export);
+		    print_variable(arg, var, ARGV(0), readonly, export);
 		} else {
 		    xerror(0, Ngt("%ls: no such variable"), arg);
-		    ok = false;
-		    continue;
 		}
 	    }
 	} while (++xoptind < argc);
     }
 
-    return ok ? Exit_SUCCESS : Exit_FAILURE;
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 }
 
 /* Prints the specified variable to the standard output.
  * This function does not print special variables whose name begins with an '='.
  * If `readonly'/`export' is true, the variable is printed only if it is
  * readonly/exported. The `name' is quoted if `is_name(name)' is not true.
- * Returns true iff successful (no error). */
-bool print_variable(
+ * An error message is printed to the standard error on error. */
+void print_variable(
 	const wchar_t *name, const variable_T *var,
 	const wchar_t *argv0, bool readonly, bool export)
 {
-    bool ok = true;
     wchar_t *qname = NULL;
 
     if (name[0] == L'=')
-	return ok;
+	return;
     if (readonly && !(var->v_type & VF_READONLY))
-	return ok;
+	return;
     if (export && !(var->v_type & VF_EXPORT))
-	return ok;
+	return;
 
     if (!is_name(name))
 	name = qname = quote_sq(name);
@@ -1663,26 +1569,22 @@ bool print_variable(
 
 print_variable:;
     wchar_t *qvalue = var->v_value ? quote_sq(var->v_value) : NULL;
+    const char *format;
     xstrbuf_T opts;
     switch (argv0[0]) {
 	case L's':
 	    assert(wcscmp(argv0, L"set") == 0);
-	    if (!qname && qvalue) {
-		if (printf("%ls=%ls\n", name, qvalue) < 0) {
+	    if (!qname && qvalue)
+		if (printf("%ls=%ls\n", name, qvalue) < 0)
 		    xerror(errno, Ngt("cannot print to standard output"));
-		    ok = false;
-		}
-	    }
 	    break;
 	case L'e':
 	case L'r':
 	    assert(wcscmp(argv0, L"export") == 0
 		    || wcscmp(argv0, L"readonly") == 0);
-	    if (printf(qvalue ? "%ls %ls=%ls\n" : "%ls %ls\n",
-			argv0, name, qvalue) < 0) {
+	    format = qvalue ? "%ls %ls=%ls\n" : "%ls %ls\n";
+	    if (printf(format, argv0, name, qvalue) < 0)
 		xerror(errno, Ngt("cannot print to standard output"));
-		ok = false;
-	    }
 	    break;
 	case L't':
 	    assert(wcscmp(argv0, L"typeset") == 0);
@@ -1693,11 +1595,9 @@ print_variable:;
 		sb_ccat(&opts, 'r');
 	    if (opts.length > 0)
 		sb_insert(&opts, 0, " -");
-	    if (printf(qvalue ? "%ls%s %ls=%ls\n" : "%ls%s %ls\n",
-			argv0, opts.contents, name, qvalue) < 0) {
+	    format = qvalue ? "%ls%s %ls=%ls\n" : "%ls%s %ls\n";
+	    if (printf(format, argv0, opts.contents, name, qvalue) < 0)
 		xerror(errno, Ngt("cannot print to standard output"));
-		ok = false;
-	    }
 	    sb_destroy(&opts);
 	    break;
 	default:
@@ -1705,7 +1605,7 @@ print_variable:;
     }
     free(qvalue);
     free(qname);
-    return ok;
+    return;
 
 print_array:
     clearerr(stdout);
@@ -1747,22 +1647,20 @@ print_array:
 	    assert(false);
     }
     free(qname);
-    if (!ferror(stdout)) {
-	return true;
-    } else {
+    if (ferror(stdout))
 	xerror(0, Ngt("cannot print to standard output"));
-	return false;
-    }
 }
 
 /* Prints the specified function to the standard output.
- * If `readonly' is true, the function is printed only if it is readonly. */
-bool print_function(
+ * If `readonly' is true, the function is printed only if it is readonly.
+ * An error message is printed to the standard error if failed to print to the
+ * standard output. */
+void print_function(
 	const wchar_t *name, const function_T *func,
 	const wchar_t *argv0, bool readonly)
 {
     if (readonly && !(func->f_type & VF_READONLY))
-	return true;
+	return;
 
     wchar_t *qname = NULL;
     if (!is_name(name))
@@ -1789,12 +1687,8 @@ bool print_function(
 	    assert(false);
     }
     free(qname);
-    if (!ferror(stdout)) {
-	return true;
-    } else {
+    if (ferror(stdout))
 	xerror(0, Ngt("cannot print to standard output"));
-	return false;
-    }
 }
 
 #if YASH_ENABLE_HELP
@@ -1876,7 +1770,6 @@ int array_builtin(int argc, void **argv)
 	return Exit_ERROR;
     }
 
-    bool ok = true;
     if (xoptind == argc) {
 	/* print all arrays */
 	if (options != 0)
@@ -1890,10 +1783,9 @@ int array_builtin(int argc, void **argv)
 	size_t count = table.count;
 	ht_destroy(&table);
 	qsort(kvs, count, sizeof *kvs, keywcscoll);
-	for (size_t i = 0; ok && i < count; i++)
+	for (size_t i = 0; yash_error_message_count == 0 && i < count; i++)
 	    if ((((variable_T *) kvs[i].value)->v_type & VF_MASK) == VF_ARRAY)
-		ok &= print_variable(
-			kvs[i].key, kvs[i].value, ARGV(0), false, false);
+		print_variable(kvs[i].key, kvs[i].value, ARGV(0), false, false);
 	free(kvs);
     } else {
 	const wchar_t *name = ARGV(xoptind++);
@@ -1903,28 +1795,27 @@ int array_builtin(int argc, void **argv)
 	}
 
 	if (options == 0) {
-	    ok = set_array(name, argc - xoptind,
+	    set_array(name, argc - xoptind,
 		    duparray(argv + xoptind, copyaswcs), SCOPE_GLOBAL);
 	} else {
 	    variable_T *array = search_array_and_check_if_changeable(name);
-	    if (!array) {
+	    if (!array)
 		return Exit_FAILURE;
-	    }
 	    switch (options) {
 		case delete:
-		    ok = array_remove_elements(
+		    array_remove_elements(
 			    array, argc - xoptind, argv + xoptind);
 		    break;
 		case insert:
 		    if (xoptind == argc)
 			goto print_usage;
-		    ok = array_insert_elements(
+		    array_insert_elements(
 			    array, argc - xoptind, argv + xoptind);
 		    break;
 		case set:
 		    if (xoptind + 2 != argc)
 			goto print_usage;
-		    ok = array_set_element(
+		    array_set_element(
 			    name, array, ARGV(xoptind), ARGV(xoptind + 1));
 		    break;
 		default:
@@ -1932,7 +1823,7 @@ int array_builtin(int argc, void **argv)
 	    }
 	}
     }
-    return ok ? Exit_SUCCESS : Exit_FAILURE;
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 }
 
 #if LONG_MAX < SIZE_MAX
@@ -1947,8 +1838,8 @@ int array_builtin(int argc, void **argv)
  * `indexwcss' is an NULL-terminated array of pointers to wide strings,
  * which are parsed as indices of elements to be removed.
  * `count' is the number of elements in `indexwcss'.
- * Returns true iff successful. */
-bool array_remove_elements(
+ * An error message is printed to the standard error on error. */
+void array_remove_elements(
 	variable_T *array, size_t count, void *const *indexwcss)
 {
     long indices[count], lastindex;
@@ -1961,7 +1852,7 @@ bool array_remove_elements(
 	const wchar_t *indexwcs = indexwcss[i];
 	if (!xwcstol(indexwcs, 10, &indices[i])) {
 	    xerror(errno, Ngt("`%ls' is not a valid integer"), indexwcs);
-	    return false;
+	    return;
 	}
     }
 
@@ -1993,7 +1884,6 @@ bool array_remove_elements(
     }
     array->v_valc = list.length;
     array->v_vals = pl_toary(&list);
-    return true;
 }
 
 int compare_long(const void *lp1, const void *lp2)
@@ -2008,8 +1898,8 @@ int compare_long(const void *lp1, const void *lp2)
  * where to insert the elements. The other strings are inserted to the array.
  * `count' is the number of strings in `values' including the first index
  * string.
- * Returns true iff successful. */
-bool array_insert_elements(
+ * An error message is printed to the standard error on error. */
+void array_insert_elements(
 	variable_T *array, size_t count, void *const *values)
 {
     long index;
@@ -2022,7 +1912,7 @@ bool array_insert_elements(
 	const wchar_t *indexword = *values;
 	if (!xwcstol(indexword, 10, &index)) {
 	    xerror(errno, Ngt("`%ls' is not a valid integer"), indexword);
-	    return false;
+	    return;
 	}
     }
     count--, values++;
@@ -2042,13 +1932,13 @@ bool array_insert_elements(
 	list.contents[uindex + i] = xwcsdup(list.contents[uindex + i]);
     array->v_valc = list.length;
     array->v_vals = pl_toary(&list);
-    return true;
 }
 
 /* Sets the value of the specified element of the array.
  * `name' is the name of the array variable.
- * `indexword' is parsed as the integer index of the element. */
-bool array_set_element(const wchar_t *name, variable_T *array,
+ * `indexword' is parsed as the integer index of the element.
+ * An error message is printed to the standard error on error. */
+void array_set_element(const wchar_t *name, variable_T *array,
 	const wchar_t *indexword, const wchar_t *value)
 {
     assert((array->v_type & VF_MASK) == VF_ARRAY);
@@ -2056,7 +1946,7 @@ bool array_set_element(const wchar_t *name, variable_T *array,
     long index;
     if (!xwcstol(indexword, 10, &index)) {
 	xerror(errno, Ngt("`%ls' is not a valid integer"), indexword);
-	return false;
+	return;
     }
 
     size_t uindex;
@@ -2076,13 +1966,12 @@ bool array_set_element(const wchar_t *name, variable_T *array,
     assert(uindex < array->v_valc);
     free(array->v_vals[uindex]);
     array->v_vals[uindex] = xwcsdup(value);
-    return true;
+    return;
 
 invalid_index:
     xerror(0,
 	    Ngt("%ls: index out of range (actual size of array `%ls' is %zu)"),
 	    indexword, name, array->v_valc);
-    return false;
 }
 
 #if YASH_ENABLE_HELP
@@ -2124,7 +2013,6 @@ int unset_builtin(int argc, void **argv)
 
     wchar_t opt;
     bool funcs = false;
-    bool err = false;
 
     xoptind = 0, xopterr = true;
     while ((opt = xgetopt_long(argv, L"fv", long_options, NULL))) {
@@ -2146,18 +2034,20 @@ int unset_builtin(int argc, void **argv)
 	const wchar_t *name = ARGV(xoptind);
 	if (wcschr(name, L'=')) {
 	    xerror(0, Ngt("`%ls': invalid name"), name);
-	    err = true;
 	    continue;
 	}
-	if (funcs ? unset_function(name) : unset_variable(name))
-	    err = true;
+	if (funcs)
+	    unset_function(name);
+	else
+	    unset_variable(name);
     }
 
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 }
 
 /* Unsets the specified function.
- * Returns true ON ERROR. */
+ * On error, an error message is printed to the standard error and TRUE is
+ * returned. */
 bool unset_function(const wchar_t *name)
 {
     kvpair_T kv = ht_remove(&functions, name);
@@ -2175,7 +2065,8 @@ bool unset_function(const wchar_t *name)
 }
 
 /* Unsets the specified variable.
- * Returns true ON ERROR. */
+ * On error, an error message is printed to the standard error and TRUE is
+ * returned. */
 bool unset_variable(const wchar_t *name)
 {
     for (environ_T *env = current_env; env; env = env->parent) {
@@ -2557,7 +2448,7 @@ int read_builtin(int argc, void **argv)
 	}
     }
 
-    bool err = false;
+    bool eof = false;
     xwcsbuf_T buf;
 
     /* read input and remove trailing newline */
@@ -2570,7 +2461,7 @@ int read_builtin(int argc, void **argv)
 	wb_remove(&buf, buf.length - 1, 1);
     } else {
 	/* no newline means the EOF was encountered */
-	err = true;
+	eof = true;
     }
 
     const wchar_t *s = buf.contents;
@@ -2591,14 +2482,14 @@ int read_builtin(int argc, void **argv)
 	if (i + 1 == list.length)
 	    trim_trailing_ifsws(list.contents[i], ifs);
 	if (!array || i + 1 < list.length)
-	    err |= !set_variable(name, list.contents[i],
-		    SCOPE_GLOBAL, shopt_allexport);
+	    set_variable(name, list.contents[i], SCOPE_GLOBAL, shopt_allexport);
 	else
-	    err |= !split_and_assign_array(name, list.contents[i], ifs, raw);
+	    split_and_assign_array(name, list.contents[i], ifs, raw);
     }
 
     pl_destroy(&list);
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    return (!eof && yash_error_message_count == 0)
+	    ? Exit_SUCCESS : Exit_FAILURE;
 
 print_usage:
     if (posixly_correct)
@@ -2608,25 +2499,74 @@ print_usage:
     return Exit_ERROR;
 }
 
-/* Reads input from the standard input and remove line continuations
- * if `noescape' is false.
+/* Reads input from the standard input and, if `noescape' is false, remove line
+ * continuations.
  * The trailing newline and all other backslashes are not removed. */
 bool read_input(xwcsbuf_T *buf, bool noescape)
 {
     if (noescape)
 	return read_line_from_stdin(buf, false);
 
-    size_t index;
     bool cont;
     bool first = true;
     do {
-	index = buf->length;
-	cont = false;
-	if (!first && is_interactive_now && isatty(STDIN_FILENO))
-	    print_prompt(2);
-	if (!read_line_from_stdin(buf, false))
-	    return false;
+	size_t index = buf->length;
+
+	if (is_interactive_now && isatty(STDIN_FILENO)) {
+	    /* if the shell is interactive and the input is from the terminal,
+	     * then print a prompt and try to use line-editing. */
+
+	    struct promptset_T prompt;
+	    if (first) {
+		prompt.main   = xwcsdup(L"");
+		prompt.right  = xwcsdup(L"");
+		prompt.styler = xwcsdup(L"");
+	    } else {
+		prompt = get_prompt(2);
+	    }
+
+#if YASH_ENABLE_LINEEDIT
+	    if (shopt_lineedit != shopt_nolineedit) {
+		wchar_t *line;
+		inputresult_T result = le_readline(prompt, &line);
+
+		if (result != INPUT_ERROR) {
+		    free_prompt(prompt);
+		    switch (result) {
+			case INPUT_OK:
+			    /* ignore lines after the first one */
+			    wcschr(line, L'\n')[1] = L'\0';
+			    wb_catfree(buf, line);
+			    /* falls thru! */
+			case INPUT_EOF:
+			    goto read;
+			case INPUT_INTERRUPTED:
+			    set_interrupted();
+			    return false;
+			case INPUT_ERROR:
+			    assert(false);
+		    }
+		}
+	    }
+#endif /* YASH_ENABLE_LINEEDIT */
+
+	    bool result2;
+	    print_prompt(prompt.main);
+	    print_prompt(prompt.styler);
+	    result2 = read_line_from_stdin(buf, false);
+	    print_prompt(PROMPT_RESET);
+	    free_prompt(prompt);
+	    if (!result2)
+		return false;
+	} else {
+	    if (!read_line_from_stdin(buf, false))
+		return false;
+	}
+#if YASH_ENABLE_LINEEDIT
+read:
+#endif
 	first = false;
+	cont = false;
 	while (index < buf->length) {
 	    if (buf->contents[index] == L'\\') {
 		if (buf->contents[index + 1] == L'\n') {
@@ -2647,15 +2587,15 @@ bool read_input(xwcsbuf_T *buf, bool noescape)
 
 /* Word-splits `values' and assigns them to the array named `name'.
  * `values' is freed in this function. */
-bool split_and_assign_array(const wchar_t *name, wchar_t *values,
+void split_and_assign_array(const wchar_t *name, wchar_t *values,
 	const wchar_t *ifs, bool raw)
 {
     plist_T list;
 
     pl_init(&list);
-    if (values[0]) {
+    if (values[0] != L'\0') {
 	const wchar_t *v = values;
-	while (*v)
+	while (*v != L'\0')
 	    pl_add(&list, split_next_field(&v, ifs, raw));
 
 	if (list.length > 0
@@ -2665,10 +2605,9 @@ bool split_and_assign_array(const wchar_t *name, wchar_t *values,
 	}
     }
 
-    bool ok = set_array(name, list.length, pl_toary(&list), SCOPE_GLOBAL);
+    set_array(name, list.length, pl_toary(&list), SCOPE_GLOBAL);
 
     free(values);
-    return ok;
 }
 
 #if YASH_ENABLE_HELP
@@ -2689,7 +2628,282 @@ const char read_help[] = Ngt(
 );
 #endif
 
-#if YASH_ENABLE_DIRSTACK
+#if !YASH_ENABLE_DIRSTACK
+
+/* options for the "pushd" builtins */
+static const struct xoption pushd_options[] = {
+    { L"default-directory", xrequired_argument, L'd', },
+    { L"logical",           xno_argument,       L'L', },
+    { L"physical",          xno_argument,       L'P', },
+#if YASH_ENABLE_HELP
+    { L"help",              xno_argument,       L'-', },
+#endif
+    { NULL, 0, 0, },
+};
+
+/* options for the "cd" and "pwd" builtins */
+const struct xoption *const cd_options  = pushd_options;
+const struct xoption *const pwd_options = pushd_options + 1;
+
+#else /* YASH_ENABLE_DIRSTACK */
+
+/* options for the "pushd" builtins */
+static const struct xoption pushd_options[] = {
+    { L"remove-duplicates", xno_argument,       L'D', },
+    { L"default-directory", xrequired_argument, L'd', },
+    { L"logical",           xno_argument,       L'L', },
+    { L"physical",          xno_argument,       L'P', },
+#if YASH_ENABLE_HELP
+    { L"help",              xno_argument,       L'-', },
+#endif
+    { NULL, 0, 0, },
+};
+
+/* options for the "cd" and "pwd" builtins */
+const struct xoption *const cd_options  = pushd_options + 1;
+const struct xoption *const pwd_options = pushd_options + 2;
+
+static variable_T *get_dirstack(void);
+static void push_dirstack(variable_T *var, wchar_t *value)
+    __attribute__((nonnull));
+static void remove_dirstack_entry_at(variable_T *var, size_t index)
+    __attribute__((nonnull));
+static void remove_dirstack_dups(variable_T *var)
+    __attribute__((nonnull));
+
+/* The "pushd" builtin.
+ *  -L: don't resolve symlinks (default)
+ *  -P: resolve symlinks
+ *  --default-directory=<dir>: go to <dir> when the operand is missing
+ *  --remove-duplicates: remove duplicate entries in the directory stack.
+ * -L and -P are mutually exclusive: the one specified last is used. */
+int pushd_builtin(int argc __attribute__((unused)), void **argv)
+{
+    bool logical = true, remove_dups = false;
+    const wchar_t *newpwd = L"+1";
+
+    wchar_t opt;
+    xoptind = 0, xopterr = true;
+    while ((opt = xgetopt_long(argv, L"-LP", pushd_options, NULL))) {
+	switch (opt) {
+	    case L'L':  logical = true;       break;
+	    case L'P':  logical = false;      break;
+	    case L'd':  newpwd = xoptarg;     break;
+	    case L'D':  remove_dups = true;   break;
+#if YASH_ENABLE_HELP
+	    case L'-':
+		return print_builtin_help(ARGV(0));
+#endif
+	    default:  print_usage:
+		fprintf(stderr, gt("Usage:  %ls [-L|-P] [dir]\n"), L"pushd");
+		return Exit_ERROR;
+	}
+    }
+
+    const wchar_t *oldpwd = getvar(L VAR_PWD);
+    if (oldpwd == NULL) {
+	xerror(0, Ngt("$PWD not set"));
+	return Exit_FAILURE;
+    }
+
+    bool printnewdir = false;
+    size_t stackindex;
+    switch (argc - xoptind) {
+	case 1:
+	    if (wcscmp(ARGV(xoptind), L"-") == 0) {
+		newpwd = getvar(L VAR_OLDPWD);
+		if (newpwd == NULL || newpwd[0] == L'\0') {
+		    xerror(0, Ngt("$OLDPWD not set"));
+		    return Exit_FAILURE;
+		}
+		printnewdir = true;
+		stackindex = SIZE_MAX;
+		break;
+	    } else {
+		newpwd = ARGV(xoptind);
+	    }
+	    /* falls thru! */
+	case 0:
+	    if (!parse_dirstack_index(newpwd, &stackindex, &newpwd, true))
+		return Exit_FAILURE;
+	    break;
+	default:
+	    goto print_usage;
+    }
+    assert(newpwd != NULL);
+
+    wchar_t *saveoldpwd = xwcsdup(oldpwd);
+    int result = change_directory(newpwd, printnewdir, logical);
+#ifndef NDEBUG
+    newpwd = NULL;  /* newpwd cannot be used anymore. */
+#endif
+    if (result != Exit_SUCCESS) {
+	free(saveoldpwd);
+	return result;
+    }
+
+    variable_T *var = get_dirstack();
+    if (var == NULL) {
+	free(saveoldpwd);
+	return Exit_FAILURE;
+    }
+
+    push_dirstack(var, saveoldpwd);
+    if (stackindex != SIZE_MAX)
+	remove_dirstack_entry_at(var, stackindex);
+    if (remove_dups)
+	remove_dirstack_dups(var);
+    return Exit_SUCCESS;
+}
+
+/* Returns the directory stack.
+ * If the stack is not yet created, it is initialized as an empty stack and
+ * returned.
+ * Fails if a non-array variable or a readonly array already exists, in which
+ * case NULL is returned. */
+variable_T *get_dirstack(void)
+{
+    variable_T *var = search_variable(L VAR_DIRSTACK);
+    if (var != NULL) {
+	if ((var->v_type & VF_MASK) != VF_ARRAY) {
+	    xerror(0, Ngt("$DIRSTACK is not an array"));
+	    return NULL;
+	} else if (var->v_type & VF_READONLY) {
+	    xerror(0, Ngt("%ls: readonly"), L VAR_DIRSTACK);
+	    return NULL;
+	}
+	return var;
+    }
+
+    void **ary = xmalloc(1 * sizeof *ary);
+    ary[0] = NULL;
+    return set_array(L VAR_DIRSTACK, 0, ary, SCOPE_GLOBAL);
+}
+
+/* Adds the specified value to the directory stack ($DIRSTACK).
+ * `var' must be the return value of a `get_dirstack' call.
+ * `value' is (virtually) freed in this function. */
+void push_dirstack(variable_T *var, wchar_t *value)
+{
+    size_t index = var->v_valc++;
+    var->v_vals = xrealloc(var->v_vals, (index + 2) * sizeof *var->v_vals);
+    var->v_vals[index] = value;
+    var->v_vals[index + 1] = NULL;
+}
+
+/* Removes the directory stack entry specified by `index'.
+ * `var' must be the directory stack array and `index' must be less than
+ * `var->v_valc'. */
+void remove_dirstack_entry_at(variable_T *var, size_t index)
+{
+    assert(index < var->v_valc);
+    free(var->v_vals[index]);
+    memmove(&var->v_vals[index], &var->v_vals[index + 1],
+	    (var->v_valc - index) * sizeof *var->v_vals);
+    var->v_valc--;
+}
+
+/* Removes directory stack entries that are the same as the current working
+ * directory ($PWD).
+ * `var' must be the return value of a `get_dirstack' call. */
+void remove_dirstack_dups(variable_T *var)
+{
+    const wchar_t *pwd = getvar(L VAR_PWD);
+    if (pwd == NULL)
+	return;
+
+    for (size_t index = var->v_valc; index-- > 0; )
+	if (wcscmp(pwd, var->v_vals[index]) == 0)
+	    remove_dirstack_entry_at(var, index);
+}
+
+#if YASH_ENABLE_HELP
+const char pushd_help[] = Ngt(
+"pushd - push directory into directory stack\n"
+"\tpushd [-L|-P] [dir]\n"
+"Changes the working directory to <dir> and appends it to the directory\n"
+"stack. Options that can be used in the \"cd\" builtin can also be used in\n"
+"the \"pushd\" builtin (-L, -P, and --default-directory=...).\n"
+"If <dir> is an integer with the plus or minus sign, it is considered a\n"
+"specific entry of the stack, which is removed from the stack and appended\n"
+"again. An integer with the plus sign specifies the nth newest entry, and\n"
+"a one with the minus sign specifies the nth oldest entry.\n"
+"If neither of <dir> and the --default-directory=... option is specified,\n"
+"\"+1\" is assumed for <dir>.\n"
+"If the --remove-duplicates option is specified, entries that are the same as\n"
+"the new working directory are removed from the stack.\n"
+);
+#endif
+
+/* The "popd" builtin. */
+int popd_builtin(int argc, void **argv)
+{
+    wchar_t opt;
+    xoptind = 0, xopterr = true;
+    while ((opt = xgetopt_long(argv, L"-", help_option, NULL))) {
+	switch (opt) {
+#if YASH_ENABLE_HELP
+	    case L'-':
+		return print_builtin_help(ARGV(0));
+#endif
+	    default:  print_usage:
+		fprintf(stderr, gt("Usage:  popd [index]\n"));
+		return Exit_ERROR;
+	}
+    }
+
+    variable_T *var = get_dirstack();
+    if (var == NULL)
+	return Exit_FAILURE;
+    if (var->v_valc == 0) {
+	xerror(0, Ngt("directory stack is empty"));
+	return Exit_FAILURE;
+    }
+
+    const wchar_t *arg;
+    switch (argc - xoptind) {
+	case 0:   arg = L"+0";          break;
+	case 1:   arg = ARGV(xoptind);  break;
+	default:  goto print_usage;
+    }
+
+    size_t stackindex;
+    const wchar_t *dummy;
+    if (!parse_dirstack_index(arg, &stackindex, &dummy, true))
+	return Exit_FAILURE;
+    if (stackindex == SIZE_MAX) {
+	xerror(0, Ngt("`%ls' is not a valid index"), arg);
+	return Exit_ERROR;
+    }
+
+    if (stackindex < var->v_valc) {
+	remove_dirstack_entry_at(var, stackindex);
+	return Exit_SUCCESS;
+    }
+
+    int result;
+    wchar_t *newpwd;
+
+    assert(var->v_valc > 0);
+    var->v_valc--;
+    newpwd = var->v_vals[var->v_valc];
+    var->v_vals[var->v_valc] = NULL;
+    result = change_directory(newpwd, true, true);
+    free(newpwd);
+    return result;
+}
+
+#if YASH_ENABLE_HELP
+const char popd_help[] = Ngt(
+"popd - pop directory from directory stack\n"
+"\tpopd [index]\n"
+"Removes the last entry of the directory stack, returning to the previous\n"
+"directory.\n"
+"If <index> is given, the entry specified by <index> is removed instead of\n"
+"the last one. An integer with the plus sign specifies the nth newest entry\n"
+"and a one with the minus sign specifies the nth oldest entry.\n"
+);
+#endif
 
 /* The "dirs" builtin, which accepts the following options:
  *  -c: clear the stack
@@ -2705,9 +2919,9 @@ int dirs_builtin(int argc, void **argv)
 	{ NULL, 0, 0, },
     };
 
-    wchar_t opt;
     bool clear = false, verbose = false;
 
+    wchar_t opt;
     xoptind = 0, xopterr = true;
     while ((opt = xgetopt_long(argv, L"-cv", long_options, NULL))) {
 	switch (opt) {
@@ -2726,58 +2940,48 @@ int dirs_builtin(int argc, void **argv)
     if (clear)
 	return unset_variable(L VAR_DIRSTACK) ? Exit_FAILURE : Exit_SUCCESS;
 
-    bool err = false;
+    variable_T *var = search_variable(L VAR_DIRSTACK);
+    bool dirvalid = (var != NULL && (var->v_type & VF_MASK) == VF_ARRAY);
+    size_t size = dirvalid ? var->v_valc : 0;
     const wchar_t *dir;
     if (xoptind < argc) {
 	/* print the specified only */
 	do {
-	    size_t index, size = get_dirstack_size();
-	    if (parse_dirstack_index(ARGV(xoptind), &index, &dir, true)) {
-		if (index != SIZE_MAX) {
-		    int r;
-		    if (verbose)
-			r = printf("+%zu\t-%zu\t%ls\n",
-				size - index, index, dir);
-		    else
-			r = printf("%ls\n", dir);
-		    if (r < 0) {
-			xerror(errno, Ngt("cannot print to standard output"));
-			err = true;
-			break;
-		    }
-		} else {
-		    xerror(0, Ngt("`%ls' is not a valid index"),
-			    ARGV(xoptind));
-		    err = true;
-		}
-	    } else {
-		err = true;
+	    size_t index;
+	    if (!parse_dirstack_index(ARGV(xoptind), &index, &dir, true))
+		continue;
+	    if (index == SIZE_MAX) {
+		xerror(0, Ngt("`%ls' is not a valid index"), ARGV(xoptind));
+		continue;
+	    }
+
+	    int r;
+	    if (verbose)
+		r = printf("+%zu\t-%zu\t%ls\n", size - index, index, dir);
+	    else
+		r = printf("%ls\n", dir);
+	    if (r < 0) {
+		xerror(errno, Ngt("cannot print to standard output"));
+		break;
 	    }
 	} while (++xoptind < argc);
     } else {
 	/* print all */
-	variable_T *var = search_variable(L VAR_DIRSTACK);
-	bool dirvalid = (var && ((var->v_type & VF_MASK) == VF_ARRAY));
-	size_t size = dirvalid ? var->v_valc : 0;
-
 	dir = getvar(L VAR_PWD);
-	if (!dir) {
+	if (dir == NULL) {
 	    xerror(0, Ngt("$PWD not set"));
-	    err = true;
 	} else {
 	    int r;
 	    if (verbose)
 		r = printf("+%zu\t-%zu\t%ls\n", (size_t) 0, size, dir);
 	    else
 		r = printf("%ls\n", dir);
-	    if (r < 0) {
+	    if (r < 0)
 		xerror(errno, Ngt("cannot print to standard output"));
-		err = true;
-	    }
 	}
 
-	if (dirvalid) {
-	    for (size_t i = var->v_valc; !err && i-- > 0; ) {
+	if (dirvalid && yash_error_message_count == 0) {
+	    for (size_t i = var->v_valc; i-- > 0; ) {
 		int r;
 		dir = var->v_vals[i];
 		if (verbose)
@@ -2786,12 +2990,13 @@ int dirs_builtin(int argc, void **argv)
 		    r = printf("%ls\n", dir);
 		if (r < 0) {
 		    xerror(errno, Ngt("cannot print to standard output"));
-		    err = true;
+		    break;
 		}
 	    }
 	}
     }
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+
+    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
 }
 
 #if YASH_ENABLE_HELP
